// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C89
// from https://raw.githubusercontent.com/fefanto/fontaudio/master/font/fontaudio.css
// for use with https://github.com/fefanto/fontaudio/raw/master/font/fontaudio.ttf
#pragma once

#define FONT_ICON_FILE_NAME_FAD "fontaudio.ttf"

#define ICON_MIN_FAD 0xf101
#define ICON_MAX_FAD 0xf19b
#define ICON_FAD_ADR "\xef\x84\x81"
#define ICON_FAD_ADSR "\xef\x84\x82"
#define ICON_FAD_AHDSR "\xef\x84\x83"
#define ICON_FAD_AR "\xef\x84\x84"
#define ICON_FAD_ARMRECORDING "\xef\x84\x85"
#define ICON_FAD_ARPCHORD "\xef\x84\x86"
#define ICON_FAD_ARPDOWN "\xef\x84\x87"
#define ICON_FAD_ARPDOWNANDUP "\xef\x84\x88"
#define ICON_FAD_ARPDOWNUP "\xef\x84\x89"
#define ICON_FAD_ARPPLAYORDER "\xef\x84\x8a"
#define ICON_FAD_ARPRANDOM "\xef\x84\x8b"
#define ICON_FAD_ARPUP "\xef\x84\x8c"
#define ICON_FAD_ARPUPANDOWN "\xef\x84\x8d"
#define ICON_FAD_ARPUPDOWN "\xef\x84\x8e"
#define ICON_FAD_ARROWS_HORZ "\xef\x84\x8f"
#define ICON_FAD_ARROWS_VERT "\xef\x84\x90"
#define ICON_FAD_AUTOMATION_2P "\xef\x84\x91"
#define ICON_FAD_AUTOMATION_3P "\xef\x84\x92"
#define ICON_FAD_AUTOMATION_4P "\xef\x84\x93"
#define ICON_FAD_BACKWARD "\xef\x84\x94"
#define ICON_FAD_BLUETOOTH "\xef\x84\x95"
#define ICON_FAD_CARET_DOWN "\xef\x84\x96"
#define ICON_FAD_CARET_LEFT "\xef\x84\x97"
#define ICON_FAD_CARET_RIGHT "\xef\x84\x98"
#define ICON_FAD_CARET_UP "\xef\x84\x99"
#define ICON_FAD_CLOSE "\xef\x84\x9a"
#define ICON_FAD_COPY "\xef\x84\x9b"
#define ICON_FAD_CPU "\xef\x84\x9c"
#define ICON_FAD_CUTTER "\xef\x84\x9d"
#define ICON_FAD_DIGITAL_COLON "\xef\x84\x9e"
#define ICON_FAD_DIGITAL_DOT "\xef\x84\x9f"
#define ICON_FAD_DIGITAL0 "\xef\x84\xa0"
#define ICON_FAD_DIGITAL1 "\xef\x84\xa1"
#define ICON_FAD_DIGITAL2 "\xef\x84\xa2"
#define ICON_FAD_DIGITAL3 "\xef\x84\xa3"
#define ICON_FAD_DIGITAL4 "\xef\x84\xa4"
#define ICON_FAD_DIGITAL5 "\xef\x84\xa5"
#define ICON_FAD_DIGITAL6 "\xef\x84\xa6"
#define ICON_FAD_DIGITAL7 "\xef\x84\xa7"
#define ICON_FAD_DIGITAL8 "\xef\x84\xa8"
#define ICON_FAD_DIGITAL9 "\xef\x84\xa9"
#define ICON_FAD_DISKIO "\xef\x84\xaa"
#define ICON_FAD_DRUMPAD "\xef\x84\xab"
#define ICON_FAD_DUPLICATE "\xef\x84\xac"
#define ICON_FAD_ERASER "\xef\x84\xad"
#define ICON_FAD_FFWD "\xef\x84\xae"
#define ICON_FAD_FILTER_BANDPASS "\xef\x84\xaf"
#define ICON_FAD_FILTER_BELL "\xef\x84\xb0"
#define ICON_FAD_FILTER_BYPASS "\xef\x84\xb1"
#define ICON_FAD_FILTER_HIGHPASS "\xef\x84\xb2"
#define ICON_FAD_FILTER_LOWPASS "\xef\x84\xb3"
#define ICON_FAD_FILTER_NOTCH "\xef\x84\xb4"
#define ICON_FAD_FILTER_REZ_HIGHPASS "\xef\x84\xb5"
#define ICON_FAD_FILTER_REZ_LOWPASS "\xef\x84\xb6"
#define ICON_FAD_FILTER_SHELVING_HI "\xef\x84\xb7"
#define ICON_FAD_FILTER_SHELVING_LO "\xef\x84\xb8"
#define ICON_FAD_FOLDBACK "\xef\x84\xb9"
#define ICON_FAD_FORWARD "\xef\x84\xba"
#define ICON_FAD_H_EXPAND "\xef\x84\xbb"
#define ICON_FAD_HARDCLIP "\xef\x84\xbc"
#define ICON_FAD_HARDCLIPCURVE "\xef\x84\xbd"
#define ICON_FAD_HEADPHONES "\xef\x84\xbe"
#define ICON_FAD_KEYBOARD "\xef\x84\xbf"
#define ICON_FAD_LOCK "\xef\x85\x80"
#define ICON_FAD_LOGO_AAX "\xef\x85\x81"
#define ICON_FAD_LOGO_ABLETONLINK "\xef\x85\x82"
#define ICON_FAD_LOGO_AU "\xef\x85\x83"
#define ICON_FAD_LOGO_AUDACITY "\xef\x85\x84"
#define ICON_FAD_LOGO_AUDIOBUS "\xef\x85\x85"
#define ICON_FAD_LOGO_CUBASE "\xef\x85\x86"
#define ICON_FAD_LOGO_FL "\xef\x85\x87"
#define ICON_FAD_LOGO_JUCE "\xef\x85\x88"
#define ICON_FAD_LOGO_LADSPA "\xef\x85\x89"
#define ICON_FAD_LOGO_LIVE "\xef\x85\x8a"
#define ICON_FAD_LOGO_LV2 "\xef\x85\x8b"
#define ICON_FAD_LOGO_PROTOOLS "\xef\x85\x8c"
#define ICON_FAD_LOGO_RACKEXT "\xef\x85\x8d"
#define ICON_FAD_LOGO_REAPER "\xef\x85\x8e"
#define ICON_FAD_LOGO_REASON "\xef\x85\x8f"
#define ICON_FAD_LOGO_REWIRE "\xef\x85\x90"
#define ICON_FAD_LOGO_STUDIOONE "\xef\x85\x91"
#define ICON_FAD_LOGO_TRACKTION "\xef\x85\x92"
#define ICON_FAD_LOGO_VST "\xef\x85\x93"
#define ICON_FAD_LOGO_WAVEFORM "\xef\x85\x94"
#define ICON_FAD_LOOP "\xef\x85\x95"
#define ICON_FAD_METRONOME "\xef\x85\x96"
#define ICON_FAD_MICROPHONE "\xef\x85\x97"
#define ICON_FAD_MIDIPLUG "\xef\x85\x98"
#define ICON_FAD_MODRANDOM "\xef\x85\x99"
#define ICON_FAD_MODSAWDOWN "\xef\x85\x9a"
#define ICON_FAD_MODSAWUP "\xef\x85\x9b"
#define ICON_FAD_MODSH "\xef\x85\x9c"
#define ICON_FAD_MODSINE "\xef\x85\x9d"
#define ICON_FAD_MODSQUARE "\xef\x85\x9e"
#define ICON_FAD_MODTRI "\xef\x85\x9f"
#define ICON_FAD_MODULARPLUG "\xef\x85\xa0"
#define ICON_FAD_MONO "\xef\x85\xa1"
#define ICON_FAD_MUTE "\xef\x85\xa2"
#define ICON_FAD_NEXT "\xef\x85\xa3"
#define ICON_FAD_OPEN "\xef\x85\xa4"
#define ICON_FAD_PASTE "\xef\x85\xa5"
#define ICON_FAD_PAUSE "\xef\x85\xa6"
#define ICON_FAD_PEN "\xef\x85\xa7"
#define ICON_FAD_PHASE "\xef\x85\xa8"
#define ICON_FAD_PLAY "\xef\x85\xa9"
#define ICON_FAD_POINTER "\xef\x85\xaa"
#define ICON_FAD_POWERSWITCH "\xef\x85\xab"
#define ICON_FAD_PRESET_A "\xef\x85\xac"
#define ICON_FAD_PRESET_AB "\xef\x85\xad"
#define ICON_FAD_PRESET_B "\xef\x85\xae"
#define ICON_FAD_PRESET_BA "\xef\x85\xaf"
#define ICON_FAD_PREV "\xef\x85\xb0"
#define ICON_FAD_PUNCH_IN "\xef\x85\xb1"
#define ICON_FAD_PUNCH_OUT "\xef\x85\xb2"
#define ICON_FAD_RAM "\xef\x85\xb3"
#define ICON_FAD_RANDOM_1DICE "\xef\x85\xb4"
#define ICON_FAD_RANDOM_2DICE "\xef\x85\xb5"
#define ICON_FAD_RECORD "\xef\x85\xb6"
#define ICON_FAD_REDO "\xef\x85\xb7"
#define ICON_FAD_REPEAT_ONE "\xef\x85\xb8"
#define ICON_FAD_REPEAT "\xef\x85\xb9"
#define ICON_FAD_REW "\xef\x85\xba"
#define ICON_FAD_ROUNDSWITCH_OFF "\xef\x85\xbb"
#define ICON_FAD_ROUNDSWITCH_ON "\xef\x85\xbc"
#define ICON_FAD_SAVE "\xef\x85\xbd"
#define ICON_FAD_SAVEAS "\xef\x85\xbe"
#define ICON_FAD_SCISSORS "\xef\x85\xbf"
#define ICON_FAD_SHUFFLE "\xef\x86\x80"
#define ICON_FAD_SLIDER_ROUND_1 "\xef\x86\x81"
#define ICON_FAD_SLIDER_ROUND_2 "\xef\x86\x82"
#define ICON_FAD_SLIDER_ROUND_3 "\xef\x86\x83"
#define ICON_FAD_SLIDERHANDLE_1 "\xef\x86\x84"
#define ICON_FAD_SLIDERHANDLE_2 "\xef\x86\x85"
#define ICON_FAD_SOFTCLIP "\xef\x86\x86"
#define ICON_FAD_SOFTCLIPCURVE "\xef\x86\x87"
#define ICON_FAD_SOLO "\xef\x86\x88"
#define ICON_FAD_SPEAKER "\xef\x86\x89"
#define ICON_FAD_SQUARESWITCH_OFF "\xef\x86\x8a"
#define ICON_FAD_SQUARESWITCH_ON "\xef\x86\x8b"
#define ICON_FAD_STEREO "\xef\x86\x8c"
#define ICON_FAD_STOP "\xef\x86\x8d"
#define ICON_FAD_THUNDERBOLT "\xef\x86\x8e"
#define ICON_FAD_TIMESELECT "\xef\x86\x8f"
#define ICON_FAD_UNDO "\xef\x86\x90"
#define ICON_FAD_UNLOCK "\xef\x86\x91"
#define ICON_FAD_USB "\xef\x86\x92"
#define ICON_FAD_V_EXPAND "\xef\x86\x93"
#define ICON_FAD_VROUNDSWITCH_OFF "\xef\x86\x94"
#define ICON_FAD_VROUNDSWITCH_ON "\xef\x86\x95"
#define ICON_FAD_VSQUARESWITCH_OFF "\xef\x86\x96"
#define ICON_FAD_VSQUARESWITCH_ON "\xef\x86\x97"
#define ICON_FAD_WAVEFORM "\xef\x86\x98"
#define ICON_FAD_XLRPLUG "\xef\x86\x99"
#define ICON_FAD_ZOOMIN "\xef\x86\x9a"
#define ICON_FAD_ZOOMOUT "\xef\x86\x9b"
