// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C++11
// from https://raw.githubusercontent.com/fefanto/fontaudio/master/font/fontaudio.css
// for use with https://github.com/fefanto/fontaudio/raw/master/font/fontaudio.ttf
#pragma once

#define FONT_ICON_FILE_NAME_FAD "fontaudio.ttf"

#define ICON_MIN_FAD 0xf101
#define ICON_MAX_FAD 0xf19b
#define ICON_FAD_ADR u8"\uf101"
#define ICON_FAD_ADSR u8"\uf102"
#define ICON_FAD_AHDSR u8"\uf103"
#define ICON_FAD_AR u8"\uf104"
#define ICON_FAD_ARMRECORDING u8"\uf105"
#define ICON_FAD_ARPCHORD u8"\uf106"
#define ICON_FAD_ARPDOWN u8"\uf107"
#define ICON_FAD_ARPDOWNANDUP u8"\uf108"
#define ICON_FAD_ARPDOWNUP u8"\uf109"
#define ICON_FAD_ARPPLAYORDER u8"\uf10a"
#define ICON_FAD_ARPRANDOM u8"\uf10b"
#define ICON_FAD_ARPUP u8"\uf10c"
#define ICON_FAD_ARPUPANDOWN u8"\uf10d"
#define ICON_FAD_ARPUPDOWN u8"\uf10e"
#define ICON_FAD_ARROWS_HORZ u8"\uf10f"
#define ICON_FAD_ARROWS_VERT u8"\uf110"
#define ICON_FAD_AUTOMATION_2P u8"\uf111"
#define ICON_FAD_AUTOMATION_3P u8"\uf112"
#define ICON_FAD_AUTOMATION_4P u8"\uf113"
#define ICON_FAD_BACKWARD u8"\uf114"
#define ICON_FAD_BLUETOOTH u8"\uf115"
#define ICON_FAD_CARET_DOWN u8"\uf116"
#define ICON_FAD_CARET_LEFT u8"\uf117"
#define ICON_FAD_CARET_RIGHT u8"\uf118"
#define ICON_FAD_CARET_UP u8"\uf119"
#define ICON_FAD_CLOSE u8"\uf11a"
#define ICON_FAD_COPY u8"\uf11b"
#define ICON_FAD_CPU u8"\uf11c"
#define ICON_FAD_CUTTER u8"\uf11d"
#define ICON_FAD_DIGITAL_COLON u8"\uf11e"
#define ICON_FAD_DIGITAL_DOT u8"\uf11f"
#define ICON_FAD_DIGITAL0 u8"\uf120"
#define ICON_FAD_DIGITAL1 u8"\uf121"
#define ICON_FAD_DIGITAL2 u8"\uf122"
#define ICON_FAD_DIGITAL3 u8"\uf123"
#define ICON_FAD_DIGITAL4 u8"\uf124"
#define ICON_FAD_DIGITAL5 u8"\uf125"
#define ICON_FAD_DIGITAL6 u8"\uf126"
#define ICON_FAD_DIGITAL7 u8"\uf127"
#define ICON_FAD_DIGITAL8 u8"\uf128"
#define ICON_FAD_DIGITAL9 u8"\uf129"
#define ICON_FAD_DISKIO u8"\uf12a"
#define ICON_FAD_DRUMPAD u8"\uf12b"
#define ICON_FAD_DUPLICATE u8"\uf12c"
#define ICON_FAD_ERASER u8"\uf12d"
#define ICON_FAD_FFWD u8"\uf12e"
#define ICON_FAD_FILTER_BANDPASS u8"\uf12f"
#define ICON_FAD_FILTER_BELL u8"\uf130"
#define ICON_FAD_FILTER_BYPASS u8"\uf131"
#define ICON_FAD_FILTER_HIGHPASS u8"\uf132"
#define ICON_FAD_FILTER_LOWPASS u8"\uf133"
#define ICON_FAD_FILTER_NOTCH u8"\uf134"
#define ICON_FAD_FILTER_REZ_HIGHPASS u8"\uf135"
#define ICON_FAD_FILTER_REZ_LOWPASS u8"\uf136"
#define ICON_FAD_FILTER_SHELVING_HI u8"\uf137"
#define ICON_FAD_FILTER_SHELVING_LO u8"\uf138"
#define ICON_FAD_FOLDBACK u8"\uf139"
#define ICON_FAD_FORWARD u8"\uf13a"
#define ICON_FAD_H_EXPAND u8"\uf13b"
#define ICON_FAD_HARDCLIP u8"\uf13c"
#define ICON_FAD_HARDCLIPCURVE u8"\uf13d"
#define ICON_FAD_HEADPHONES u8"\uf13e"
#define ICON_FAD_KEYBOARD u8"\uf13f"
#define ICON_FAD_LOCK u8"\uf140"
#define ICON_FAD_LOGO_AAX u8"\uf141"
#define ICON_FAD_LOGO_ABLETONLINK u8"\uf142"
#define ICON_FAD_LOGO_AU u8"\uf143"
#define ICON_FAD_LOGO_AUDACITY u8"\uf144"
#define ICON_FAD_LOGO_AUDIOBUS u8"\uf145"
#define ICON_FAD_LOGO_CUBASE u8"\uf146"
#define ICON_FAD_LOGO_FL u8"\uf147"
#define ICON_FAD_LOGO_JUCE u8"\uf148"
#define ICON_FAD_LOGO_LADSPA u8"\uf149"
#define ICON_FAD_LOGO_LIVE u8"\uf14a"
#define ICON_FAD_LOGO_LV2 u8"\uf14b"
#define ICON_FAD_LOGO_PROTOOLS u8"\uf14c"
#define ICON_FAD_LOGO_RACKEXT u8"\uf14d"
#define ICON_FAD_LOGO_REAPER u8"\uf14e"
#define ICON_FAD_LOGO_REASON u8"\uf14f"
#define ICON_FAD_LOGO_REWIRE u8"\uf150"
#define ICON_FAD_LOGO_STUDIOONE u8"\uf151"
#define ICON_FAD_LOGO_TRACKTION u8"\uf152"
#define ICON_FAD_LOGO_VST u8"\uf153"
#define ICON_FAD_LOGO_WAVEFORM u8"\uf154"
#define ICON_FAD_LOOP u8"\uf155"
#define ICON_FAD_METRONOME u8"\uf156"
#define ICON_FAD_MICROPHONE u8"\uf157"
#define ICON_FAD_MIDIPLUG u8"\uf158"
#define ICON_FAD_MODRANDOM u8"\uf159"
#define ICON_FAD_MODSAWDOWN u8"\uf15a"
#define ICON_FAD_MODSAWUP u8"\uf15b"
#define ICON_FAD_MODSH u8"\uf15c"
#define ICON_FAD_MODSINE u8"\uf15d"
#define ICON_FAD_MODSQUARE u8"\uf15e"
#define ICON_FAD_MODTRI u8"\uf15f"
#define ICON_FAD_MODULARPLUG u8"\uf160"
#define ICON_FAD_MONO u8"\uf161"
#define ICON_FAD_MUTE u8"\uf162"
#define ICON_FAD_NEXT u8"\uf163"
#define ICON_FAD_OPEN u8"\uf164"
#define ICON_FAD_PASTE u8"\uf165"
#define ICON_FAD_PAUSE u8"\uf166"
#define ICON_FAD_PEN u8"\uf167"
#define ICON_FAD_PHASE u8"\uf168"
#define ICON_FAD_PLAY u8"\uf169"
#define ICON_FAD_POINTER u8"\uf16a"
#define ICON_FAD_POWERSWITCH u8"\uf16b"
#define ICON_FAD_PRESET_A u8"\uf16c"
#define ICON_FAD_PRESET_AB u8"\uf16d"
#define ICON_FAD_PRESET_B u8"\uf16e"
#define ICON_FAD_PRESET_BA u8"\uf16f"
#define ICON_FAD_PREV u8"\uf170"
#define ICON_FAD_PUNCH_IN u8"\uf171"
#define ICON_FAD_PUNCH_OUT u8"\uf172"
#define ICON_FAD_RAM u8"\uf173"
#define ICON_FAD_RANDOM_1DICE u8"\uf174"
#define ICON_FAD_RANDOM_2DICE u8"\uf175"
#define ICON_FAD_RECORD u8"\uf176"
#define ICON_FAD_REDO u8"\uf177"
#define ICON_FAD_REPEAT_ONE u8"\uf178"
#define ICON_FAD_REPEAT u8"\uf179"
#define ICON_FAD_REW u8"\uf17a"
#define ICON_FAD_ROUNDSWITCH_OFF u8"\uf17b"
#define ICON_FAD_ROUNDSWITCH_ON u8"\uf17c"
#define ICON_FAD_SAVE u8"\uf17d"
#define ICON_FAD_SAVEAS u8"\uf17e"
#define ICON_FAD_SCISSORS u8"\uf17f"
#define ICON_FAD_SHUFFLE u8"\uf180"
#define ICON_FAD_SLIDER_ROUND_1 u8"\uf181"
#define ICON_FAD_SLIDER_ROUND_2 u8"\uf182"
#define ICON_FAD_SLIDER_ROUND_3 u8"\uf183"
#define ICON_FAD_SLIDERHANDLE_1 u8"\uf184"
#define ICON_FAD_SLIDERHANDLE_2 u8"\uf185"
#define ICON_FAD_SOFTCLIP u8"\uf186"
#define ICON_FAD_SOFTCLIPCURVE u8"\uf187"
#define ICON_FAD_SOLO u8"\uf188"
#define ICON_FAD_SPEAKER u8"\uf189"
#define ICON_FAD_SQUARESWITCH_OFF u8"\uf18a"
#define ICON_FAD_SQUARESWITCH_ON u8"\uf18b"
#define ICON_FAD_STEREO u8"\uf18c"
#define ICON_FAD_STOP u8"\uf18d"
#define ICON_FAD_THUNDERBOLT u8"\uf18e"
#define ICON_FAD_TIMESELECT u8"\uf18f"
#define ICON_FAD_UNDO u8"\uf190"
#define ICON_FAD_UNLOCK u8"\uf191"
#define ICON_FAD_USB u8"\uf192"
#define ICON_FAD_V_EXPAND u8"\uf193"
#define ICON_FAD_VROUNDSWITCH_OFF u8"\uf194"
#define ICON_FAD_VROUNDSWITCH_ON u8"\uf195"
#define ICON_FAD_VSQUARESWITCH_OFF u8"\uf196"
#define ICON_FAD_VSQUARESWITCH_ON u8"\uf197"
#define ICON_FAD_WAVEFORM u8"\uf198"
#define ICON_FAD_XLRPLUG u8"\uf199"
#define ICON_FAD_ZOOMIN u8"\uf19a"
#define ICON_FAD_ZOOMOUT u8"\uf19b"
