/**
 * Furnace Tracker - multi-system chiptune tracker
 * Copyright (C) 2021-2024 tildearrow and contributors
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#include "../effect.h"

class DivEffectDummy: public DivEffect {
  public:
    void acquire(float** in, float** out, size_t len);
    void reset();
    int getInputCount();
    int getOutputCount();
    const char* getParams();
    size_t getParamCount();
    bool load(unsigned short version, const unsigned char* data, size_t len);
    unsigned char* save(unsigned short* version, size_t* len);
    bool init(DivEngine* parent, double rate, unsigned short version, const unsigned char* data, size_t len);
    void quit();
};