// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C#
// from icons.yml
// for use with fa-light-300.ttf, fa-regular-400.ttf, fa-solid-900.ttf
namespace IconFonts
{
    public class FontAwesome5Pro
    {
        public const string FontIconFileName = "fa-light-300.ttf";
        public const string FontIconFileName = "fa-regular-400.ttf";
        public const string FontIconFileName = "fa-solid-900.ttf";

        public const int IconMin = 0xe000;
        public const int IconMax = 0xf8ff;
        public const string Abacus = "\uf640";
        public const string Acorn = "\uf6ae";
        public const string Ad = "\uf641";
        public const string AddressBook = "\uf2b9";
        public const string AddressCard = "\uf2bb";
        public const string Adjust = "\uf042";
        public const string AirConditioner = "\uf8f4";
        public const string AirFreshener = "\uf5d0";
        public const string AlarmClock = "\uf34e";
        public const string AlarmExclamation = "\uf843";
        public const string AlarmPlus = "\uf844";
        public const string AlarmSnooze = "\uf845";
        public const string Album = "\uf89f";
        public const string AlbumCollection = "\uf8a0";
        public const string Alicorn = "\uf6b0";
        public const string Alien = "\uf8f5";
        public const string AlienMonster = "\uf8f6";
        public const string AlignCenter = "\uf037";
        public const string AlignJustify = "\uf039";
        public const string AlignLeft = "\uf036";
        public const string AlignRight = "\uf038";
        public const string AlignSlash = "\uf846";
        public const string Allergies = "\uf461";
        public const string Ambulance = "\uf0f9";
        public const string AmericanSignLanguageInterpreting = "\uf2a3";
        public const string AmpGuitar = "\uf8a1";
        public const string Analytics = "\uf643";
        public const string Anchor = "\uf13d";
        public const string Angel = "\uf779";
        public const string AngleDoubleDown = "\uf103";
        public const string AngleDoubleLeft = "\uf100";
        public const string AngleDoubleRight = "\uf101";
        public const string AngleDoubleUp = "\uf102";
        public const string AngleDown = "\uf107";
        public const string AngleLeft = "\uf104";
        public const string AngleRight = "\uf105";
        public const string AngleUp = "\uf106";
        public const string Angry = "\uf556";
        public const string Ankh = "\uf644";
        public const string AppleAlt = "\uf5d1";
        public const string AppleCrate = "\uf6b1";
        public const string Archive = "\uf187";
        public const string Archway = "\uf557";
        public const string ArrowAltCircleDown = "\uf358";
        public const string ArrowAltCircleLeft = "\uf359";
        public const string ArrowAltCircleRight = "\uf35a";
        public const string ArrowAltCircleUp = "\uf35b";
        public const string ArrowAltDown = "\uf354";
        public const string ArrowAltFromBottom = "\uf346";
        public const string ArrowAltFromLeft = "\uf347";
        public const string ArrowAltFromRight = "\uf348";
        public const string ArrowAltFromTop = "\uf349";
        public const string ArrowAltLeft = "\uf355";
        public const string ArrowAltRight = "\uf356";
        public const string ArrowAltSquareDown = "\uf350";
        public const string ArrowAltSquareLeft = "\uf351";
        public const string ArrowAltSquareRight = "\uf352";
        public const string ArrowAltSquareUp = "\uf353";
        public const string ArrowAltToBottom = "\uf34a";
        public const string ArrowAltToLeft = "\uf34b";
        public const string ArrowAltToRight = "\uf34c";
        public const string ArrowAltToTop = "\uf34d";
        public const string ArrowAltUp = "\uf357";
        public const string ArrowCircleDown = "\uf0ab";
        public const string ArrowCircleLeft = "\uf0a8";
        public const string ArrowCircleRight = "\uf0a9";
        public const string ArrowCircleUp = "\uf0aa";
        public const string ArrowDown = "\uf063";
        public const string ArrowFromBottom = "\uf342";
        public const string ArrowFromLeft = "\uf343";
        public const string ArrowFromRight = "\uf344";
        public const string ArrowFromTop = "\uf345";
        public const string ArrowLeft = "\uf060";
        public const string ArrowRight = "\uf061";
        public const string ArrowSquareDown = "\uf339";
        public const string ArrowSquareLeft = "\uf33a";
        public const string ArrowSquareRight = "\uf33b";
        public const string ArrowSquareUp = "\uf33c";
        public const string ArrowToBottom = "\uf33d";
        public const string ArrowToLeft = "\uf33e";
        public const string ArrowToRight = "\uf340";
        public const string ArrowToTop = "\uf341";
        public const string ArrowUp = "\uf062";
        public const string Arrows = "\uf047";
        public const string ArrowsAlt = "\uf0b2";
        public const string ArrowsAltH = "\uf337";
        public const string ArrowsAltV = "\uf338";
        public const string ArrowsH = "\uf07e";
        public const string ArrowsV = "\uf07d";
        public const string AssistiveListeningSystems = "\uf2a2";
        public const string Asterisk = "\uf069";
        public const string At = "\uf1fa";
        public const string Atlas = "\uf558";
        public const string Atom = "\uf5d2";
        public const string AtomAlt = "\uf5d3";
        public const string AudioDescription = "\uf29e";
        public const string Award = "\uf559";
        public const string Axe = "\uf6b2";
        public const string AxeBattle = "\uf6b3";
        public const string Baby = "\uf77c";
        public const string BabyCarriage = "\uf77d";
        public const string Backpack = "\uf5d4";
        public const string Backspace = "\uf55a";
        public const string Backward = "\uf04a";
        public const string Bacon = "\uf7e5";
        public const string Bacteria = "\ue059";
        public const string Bacterium = "\ue05a";
        public const string Badge = "\uf335";
        public const string BadgeCheck = "\uf336";
        public const string BadgeDollar = "\uf645";
        public const string BadgePercent = "\uf646";
        public const string BadgeSheriff = "\uf8a2";
        public const string BadgerHoney = "\uf6b4";
        public const string BagsShopping = "\uf847";
        public const string Bahai = "\uf666";
        public const string BalanceScale = "\uf24e";
        public const string BalanceScaleLeft = "\uf515";
        public const string BalanceScaleRight = "\uf516";
        public const string BallPile = "\uf77e";
        public const string Ballot = "\uf732";
        public const string BallotCheck = "\uf733";
        public const string Ban = "\uf05e";
        public const string BandAid = "\uf462";
        public const string Banjo = "\uf8a3";
        public const string Barcode = "\uf02a";
        public const string BarcodeAlt = "\uf463";
        public const string BarcodeRead = "\uf464";
        public const string BarcodeScan = "\uf465";
        public const string Bars = "\uf0c9";
        public const string Baseball = "\uf432";
        public const string BaseballBall = "\uf433";
        public const string BasketballBall = "\uf434";
        public const string BasketballHoop = "\uf435";
        public const string Bat = "\uf6b5";
        public const string Bath = "\uf2cd";
        public const string BatteryBolt = "\uf376";
        public const string BatteryEmpty = "\uf244";
        public const string BatteryFull = "\uf240";
        public const string BatteryHalf = "\uf242";
        public const string BatteryQuarter = "\uf243";
        public const string BatterySlash = "\uf377";
        public const string BatteryThreeQuarters = "\uf241";
        public const string Bed = "\uf236";
        public const string BedAlt = "\uf8f7";
        public const string BedBunk = "\uf8f8";
        public const string BedEmpty = "\uf8f9";
        public const string Beer = "\uf0fc";
        public const string Bell = "\uf0f3";
        public const string BellExclamation = "\uf848";
        public const string BellOn = "\uf8fa";
        public const string BellPlus = "\uf849";
        public const string BellSchool = "\uf5d5";
        public const string BellSchoolSlash = "\uf5d6";
        public const string BellSlash = "\uf1f6";
        public const string Bells = "\uf77f";
        public const string Betamax = "\uf8a4";
        public const string BezierCurve = "\uf55b";
        public const string Bible = "\uf647";
        public const string Bicycle = "\uf206";
        public const string Biking = "\uf84a";
        public const string BikingMountain = "\uf84b";
        public const string Binoculars = "\uf1e5";
        public const string Biohazard = "\uf780";
        public const string BirthdayCake = "\uf1fd";
        public const string Blanket = "\uf498";
        public const string Blender = "\uf517";
        public const string BlenderPhone = "\uf6b6";
        public const string Blind = "\uf29d";
        public const string Blinds = "\uf8fb";
        public const string BlindsOpen = "\uf8fc";
        public const string BlindsRaised = "\uf8fd";
        public const string Blog = "\uf781";
        public const string Bold = "\uf032";
        public const string Bolt = "\uf0e7";
        public const string Bomb = "\uf1e2";
        public const string Bone = "\uf5d7";
        public const string BoneBreak = "\uf5d8";
        public const string Bong = "\uf55c";
        public const string Book = "\uf02d";
        public const string BookAlt = "\uf5d9";
        public const string BookDead = "\uf6b7";
        public const string BookHeart = "\uf499";
        public const string BookMedical = "\uf7e6";
        public const string BookOpen = "\uf518";
        public const string BookReader = "\uf5da";
        public const string BookSpells = "\uf6b8";
        public const string BookUser = "\uf7e7";
        public const string Bookmark = "\uf02e";
        public const string Books = "\uf5db";
        public const string BooksMedical = "\uf7e8";
        public const string Boombox = "\uf8a5";
        public const string Boot = "\uf782";
        public const string BoothCurtain = "\uf734";
        public const string BorderAll = "\uf84c";
        public const string BorderBottom = "\uf84d";
        public const string BorderCenterH = "\uf89c";
        public const string BorderCenterV = "\uf89d";
        public const string BorderInner = "\uf84e";
        public const string BorderLeft = "\uf84f";
        public const string BorderNone = "\uf850";
        public const string BorderOuter = "\uf851";
        public const string BorderRight = "\uf852";
        public const string BorderStyle = "\uf853";
        public const string BorderStyleAlt = "\uf854";
        public const string BorderTop = "\uf855";
        public const string BowArrow = "\uf6b9";
        public const string BowlingBall = "\uf436";
        public const string BowlingPins = "\uf437";
        public const string Box = "\uf466";
        public const string BoxAlt = "\uf49a";
        public const string BoxBallot = "\uf735";
        public const string BoxCheck = "\uf467";
        public const string BoxFragile = "\uf49b";
        public const string BoxFull = "\uf49c";
        public const string BoxHeart = "\uf49d";
        public const string BoxOpen = "\uf49e";
        public const string BoxTissue = "\ue05b";
        public const string BoxUp = "\uf49f";
        public const string BoxUsd = "\uf4a0";
        public const string Boxes = "\uf468";
        public const string BoxesAlt = "\uf4a1";
        public const string BoxingGlove = "\uf438";
        public const string Brackets = "\uf7e9";
        public const string BracketsCurly = "\uf7ea";
        public const string Braille = "\uf2a1";
        public const string Brain = "\uf5dc";
        public const string BreadLoaf = "\uf7eb";
        public const string BreadSlice = "\uf7ec";
        public const string Briefcase = "\uf0b1";
        public const string BriefcaseMedical = "\uf469";
        public const string BringForward = "\uf856";
        public const string BringFront = "\uf857";
        public const string BroadcastTower = "\uf519";
        public const string Broom = "\uf51a";
        public const string Browser = "\uf37e";
        public const string Brush = "\uf55d";
        public const string Bug = "\uf188";
        public const string Building = "\uf1ad";
        public const string Bullhorn = "\uf0a1";
        public const string Bullseye = "\uf140";
        public const string BullseyeArrow = "\uf648";
        public const string BullseyePointer = "\uf649";
        public const string BurgerSoda = "\uf858";
        public const string Burn = "\uf46a";
        public const string Burrito = "\uf7ed";
        public const string Bus = "\uf207";
        public const string BusAlt = "\uf55e";
        public const string BusSchool = "\uf5dd";
        public const string BusinessTime = "\uf64a";
        public const string CabinetFiling = "\uf64b";
        public const string Cactus = "\uf8a7";
        public const string Calculator = "\uf1ec";
        public const string CalculatorAlt = "\uf64c";
        public const string Calendar = "\uf133";
        public const string CalendarAlt = "\uf073";
        public const string CalendarCheck = "\uf274";
        public const string CalendarDay = "\uf783";
        public const string CalendarEdit = "\uf333";
        public const string CalendarExclamation = "\uf334";
        public const string CalendarMinus = "\uf272";
        public const string CalendarPlus = "\uf271";
        public const string CalendarStar = "\uf736";
        public const string CalendarTimes = "\uf273";
        public const string CalendarWeek = "\uf784";
        public const string Camcorder = "\uf8a8";
        public const string Camera = "\uf030";
        public const string CameraAlt = "\uf332";
        public const string CameraHome = "\uf8fe";
        public const string CameraMovie = "\uf8a9";
        public const string CameraPolaroid = "\uf8aa";
        public const string CameraRetro = "\uf083";
        public const string Campfire = "\uf6ba";
        public const string Campground = "\uf6bb";
        public const string CandleHolder = "\uf6bc";
        public const string CandyCane = "\uf786";
        public const string CandyCorn = "\uf6bd";
        public const string Cannabis = "\uf55f";
        public const string Capsules = "\uf46b";
        public const string Car = "\uf1b9";
        public const string CarAlt = "\uf5de";
        public const string CarBattery = "\uf5df";
        public const string CarBuilding = "\uf859";
        public const string CarBump = "\uf5e0";
        public const string CarBus = "\uf85a";
        public const string CarCrash = "\uf5e1";
        public const string CarGarage = "\uf5e2";
        public const string CarMechanic = "\uf5e3";
        public const string CarSide = "\uf5e4";
        public const string CarTilt = "\uf5e5";
        public const string CarWash = "\uf5e6";
        public const string Caravan = "\uf8ff";
        public const string CaravanAlt = "\ue000";
        public const string CaretCircleDown = "\uf32d";
        public const string CaretCircleLeft = "\uf32e";
        public const string CaretCircleRight = "\uf330";
        public const string CaretCircleUp = "\uf331";
        public const string CaretDown = "\uf0d7";
        public const string CaretLeft = "\uf0d9";
        public const string CaretRight = "\uf0da";
        public const string CaretSquareDown = "\uf150";
        public const string CaretSquareLeft = "\uf191";
        public const string CaretSquareRight = "\uf152";
        public const string CaretSquareUp = "\uf151";
        public const string CaretUp = "\uf0d8";
        public const string Carrot = "\uf787";
        public const string Cars = "\uf85b";
        public const string CartArrowDown = "\uf218";
        public const string CartPlus = "\uf217";
        public const string CashRegister = "\uf788";
        public const string CassetteTape = "\uf8ab";
        public const string Cat = "\uf6be";
        public const string CatSpace = "\ue001";
        public const string Cauldron = "\uf6bf";
        public const string Cctv = "\uf8ac";
        public const string Certificate = "\uf0a3";
        public const string Chair = "\uf6c0";
        public const string ChairOffice = "\uf6c1";
        public const string Chalkboard = "\uf51b";
        public const string ChalkboardTeacher = "\uf51c";
        public const string ChargingStation = "\uf5e7";
        public const string ChartArea = "\uf1fe";
        public const string ChartBar = "\uf080";
        public const string ChartLine = "\uf201";
        public const string ChartLineDown = "\uf64d";
        public const string ChartNetwork = "\uf78a";
        public const string ChartPie = "\uf200";
        public const string ChartPieAlt = "\uf64e";
        public const string ChartScatter = "\uf7ee";
        public const string Check = "\uf00c";
        public const string CheckCircle = "\uf058";
        public const string CheckDouble = "\uf560";
        public const string CheckSquare = "\uf14a";
        public const string Cheese = "\uf7ef";
        public const string CheeseSwiss = "\uf7f0";
        public const string Cheeseburger = "\uf7f1";
        public const string Chess = "\uf439";
        public const string ChessBishop = "\uf43a";
        public const string ChessBishopAlt = "\uf43b";
        public const string ChessBoard = "\uf43c";
        public const string ChessClock = "\uf43d";
        public const string ChessClockAlt = "\uf43e";
        public const string ChessKing = "\uf43f";
        public const string ChessKingAlt = "\uf440";
        public const string ChessKnight = "\uf441";
        public const string ChessKnightAlt = "\uf442";
        public const string ChessPawn = "\uf443";
        public const string ChessPawnAlt = "\uf444";
        public const string ChessQueen = "\uf445";
        public const string ChessQueenAlt = "\uf446";
        public const string ChessRook = "\uf447";
        public const string ChessRookAlt = "\uf448";
        public const string ChevronCircleDown = "\uf13a";
        public const string ChevronCircleLeft = "\uf137";
        public const string ChevronCircleRight = "\uf138";
        public const string ChevronCircleUp = "\uf139";
        public const string ChevronDoubleDown = "\uf322";
        public const string ChevronDoubleLeft = "\uf323";
        public const string ChevronDoubleRight = "\uf324";
        public const string ChevronDoubleUp = "\uf325";
        public const string ChevronDown = "\uf078";
        public const string ChevronLeft = "\uf053";
        public const string ChevronRight = "\uf054";
        public const string ChevronSquareDown = "\uf329";
        public const string ChevronSquareLeft = "\uf32a";
        public const string ChevronSquareRight = "\uf32b";
        public const string ChevronSquareUp = "\uf32c";
        public const string ChevronUp = "\uf077";
        public const string Child = "\uf1ae";
        public const string Chimney = "\uf78b";
        public const string Church = "\uf51d";
        public const string Circle = "\uf111";
        public const string CircleNotch = "\uf1ce";
        public const string City = "\uf64f";
        public const string Clarinet = "\uf8ad";
        public const string ClawMarks = "\uf6c2";
        public const string ClinicMedical = "\uf7f2";
        public const string Clipboard = "\uf328";
        public const string ClipboardCheck = "\uf46c";
        public const string ClipboardList = "\uf46d";
        public const string ClipboardListCheck = "\uf737";
        public const string ClipboardPrescription = "\uf5e8";
        public const string ClipboardUser = "\uf7f3";
        public const string Clock = "\uf017";
        public const string Clone = "\uf24d";
        public const string ClosedCaptioning = "\uf20a";
        public const string Cloud = "\uf0c2";
        public const string CloudDownload = "\uf0ed";
        public const string CloudDownloadAlt = "\uf381";
        public const string CloudDrizzle = "\uf738";
        public const string CloudHail = "\uf739";
        public const string CloudHailMixed = "\uf73a";
        public const string CloudMeatball = "\uf73b";
        public const string CloudMoon = "\uf6c3";
        public const string CloudMoonRain = "\uf73c";
        public const string CloudMusic = "\uf8ae";
        public const string CloudRain = "\uf73d";
        public const string CloudRainbow = "\uf73e";
        public const string CloudShowers = "\uf73f";
        public const string CloudShowersHeavy = "\uf740";
        public const string CloudSleet = "\uf741";
        public const string CloudSnow = "\uf742";
        public const string CloudSun = "\uf6c4";
        public const string CloudSunRain = "\uf743";
        public const string CloudUpload = "\uf0ee";
        public const string CloudUploadAlt = "\uf382";
        public const string Clouds = "\uf744";
        public const string CloudsMoon = "\uf745";
        public const string CloudsSun = "\uf746";
        public const string Club = "\uf327";
        public const string Cocktail = "\uf561";
        public const string Code = "\uf121";
        public const string CodeBranch = "\uf126";
        public const string CodeCommit = "\uf386";
        public const string CodeMerge = "\uf387";
        public const string Coffee = "\uf0f4";
        public const string CoffeePot = "\ue002";
        public const string CoffeeTogo = "\uf6c5";
        public const string Coffin = "\uf6c6";
        public const string CoffinCross = "\ue051";
        public const string Cog = "\uf013";
        public const string Cogs = "\uf085";
        public const string Coin = "\uf85c";
        public const string Coins = "\uf51e";
        public const string Columns = "\uf0db";
        public const string Comet = "\ue003";
        public const string Comment = "\uf075";
        public const string CommentAlt = "\uf27a";
        public const string CommentAltCheck = "\uf4a2";
        public const string CommentAltDollar = "\uf650";
        public const string CommentAltDots = "\uf4a3";
        public const string CommentAltEdit = "\uf4a4";
        public const string CommentAltExclamation = "\uf4a5";
        public const string CommentAltLines = "\uf4a6";
        public const string CommentAltMedical = "\uf7f4";
        public const string CommentAltMinus = "\uf4a7";
        public const string CommentAltMusic = "\uf8af";
        public const string CommentAltPlus = "\uf4a8";
        public const string CommentAltSlash = "\uf4a9";
        public const string CommentAltSmile = "\uf4aa";
        public const string CommentAltTimes = "\uf4ab";
        public const string CommentCheck = "\uf4ac";
        public const string CommentDollar = "\uf651";
        public const string CommentDots = "\uf4ad";
        public const string CommentEdit = "\uf4ae";
        public const string CommentExclamation = "\uf4af";
        public const string CommentLines = "\uf4b0";
        public const string CommentMedical = "\uf7f5";
        public const string CommentMinus = "\uf4b1";
        public const string CommentMusic = "\uf8b0";
        public const string CommentPlus = "\uf4b2";
        public const string CommentSlash = "\uf4b3";
        public const string CommentSmile = "\uf4b4";
        public const string CommentTimes = "\uf4b5";
        public const string Comments = "\uf086";
        public const string CommentsAlt = "\uf4b6";
        public const string CommentsAltDollar = "\uf652";
        public const string CommentsDollar = "\uf653";
        public const string CompactDisc = "\uf51f";
        public const string Compass = "\uf14e";
        public const string CompassSlash = "\uf5e9";
        public const string Compress = "\uf066";
        public const string CompressAlt = "\uf422";
        public const string CompressArrowsAlt = "\uf78c";
        public const string CompressWide = "\uf326";
        public const string ComputerClassic = "\uf8b1";
        public const string ComputerSpeaker = "\uf8b2";
        public const string ConciergeBell = "\uf562";
        public const string Construction = "\uf85d";
        public const string ContainerStorage = "\uf4b7";
        public const string ConveyorBelt = "\uf46e";
        public const string ConveyorBeltAlt = "\uf46f";
        public const string Cookie = "\uf563";
        public const string CookieBite = "\uf564";
        public const string Copy = "\uf0c5";
        public const string Copyright = "\uf1f9";
        public const string Corn = "\uf6c7";
        public const string Couch = "\uf4b8";
        public const string Cow = "\uf6c8";
        public const string Cowbell = "\uf8b3";
        public const string CowbellMore = "\uf8b4";
        public const string CreditCard = "\uf09d";
        public const string CreditCardBlank = "\uf389";
        public const string CreditCardFront = "\uf38a";
        public const string Cricket = "\uf449";
        public const string Croissant = "\uf7f6";
        public const string Crop = "\uf125";
        public const string CropAlt = "\uf565";
        public const string Cross = "\uf654";
        public const string Crosshairs = "\uf05b";
        public const string Crow = "\uf520";
        public const string Crown = "\uf521";
        public const string Crutch = "\uf7f7";
        public const string Crutches = "\uf7f8";
        public const string Cube = "\uf1b2";
        public const string Cubes = "\uf1b3";
        public const string Curling = "\uf44a";
        public const string Cut = "\uf0c4";
        public const string Dagger = "\uf6cb";
        public const string Database = "\uf1c0";
        public const string Deaf = "\uf2a4";
        public const string Debug = "\uf7f9";
        public const string Deer = "\uf78e";
        public const string DeerRudolph = "\uf78f";
        public const string Democrat = "\uf747";
        public const string Desktop = "\uf108";
        public const string DesktopAlt = "\uf390";
        public const string Dewpoint = "\uf748";
        public const string Dharmachakra = "\uf655";
        public const string Diagnoses = "\uf470";
        public const string Diamond = "\uf219";
        public const string Dice = "\uf522";
        public const string DiceD10 = "\uf6cd";
        public const string DiceD12 = "\uf6ce";
        public const string DiceD20 = "\uf6cf";
        public const string DiceD4 = "\uf6d0";
        public const string DiceD6 = "\uf6d1";
        public const string DiceD8 = "\uf6d2";
        public const string DiceFive = "\uf523";
        public const string DiceFour = "\uf524";
        public const string DiceOne = "\uf525";
        public const string DiceSix = "\uf526";
        public const string DiceThree = "\uf527";
        public const string DiceTwo = "\uf528";
        public const string Digging = "\uf85e";
        public const string DigitalTachograph = "\uf566";
        public const string Diploma = "\uf5ea";
        public const string Directions = "\uf5eb";
        public const string DiscDrive = "\uf8b5";
        public const string Disease = "\uf7fa";
        public const string Divide = "\uf529";
        public const string Dizzy = "\uf567";
        public const string Dna = "\uf471";
        public const string DoNotEnter = "\uf5ec";
        public const string Dog = "\uf6d3";
        public const string DogLeashed = "\uf6d4";
        public const string DollarSign = "\uf155";
        public const string Dolly = "\uf472";
        public const string DollyEmpty = "\uf473";
        public const string DollyFlatbed = "\uf474";
        public const string DollyFlatbedAlt = "\uf475";
        public const string DollyFlatbedEmpty = "\uf476";
        public const string Donate = "\uf4b9";
        public const string DoorClosed = "\uf52a";
        public const string DoorOpen = "\uf52b";
        public const string DotCircle = "\uf192";
        public const string Dove = "\uf4ba";
        public const string Download = "\uf019";
        public const string DraftingCompass = "\uf568";
        public const string Dragon = "\uf6d5";
        public const string DrawCircle = "\uf5ed";
        public const string DrawPolygon = "\uf5ee";
        public const string DrawSquare = "\uf5ef";
        public const string Dreidel = "\uf792";
        public const string Drone = "\uf85f";
        public const string DroneAlt = "\uf860";
        public const string Drum = "\uf569";
        public const string DrumSteelpan = "\uf56a";
        public const string Drumstick = "\uf6d6";
        public const string DrumstickBite = "\uf6d7";
        public const string Dryer = "\uf861";
        public const string DryerAlt = "\uf862";
        public const string Duck = "\uf6d8";
        public const string Dumbbell = "\uf44b";
        public const string Dumpster = "\uf793";
        public const string DumpsterFire = "\uf794";
        public const string Dungeon = "\uf6d9";
        public const string Ear = "\uf5f0";
        public const string EarMuffs = "\uf795";
        public const string Eclipse = "\uf749";
        public const string EclipseAlt = "\uf74a";
        public const string Edit = "\uf044";
        public const string Egg = "\uf7fb";
        public const string EggFried = "\uf7fc";
        public const string Eject = "\uf052";
        public const string Elephant = "\uf6da";
        public const string EllipsisH = "\uf141";
        public const string EllipsisHAlt = "\uf39b";
        public const string EllipsisV = "\uf142";
        public const string EllipsisVAlt = "\uf39c";
        public const string EmptySet = "\uf656";
        public const string EngineWarning = "\uf5f2";
        public const string Envelope = "\uf0e0";
        public const string EnvelopeOpen = "\uf2b6";
        public const string EnvelopeOpenDollar = "\uf657";
        public const string EnvelopeOpenText = "\uf658";
        public const string EnvelopeSquare = "\uf199";
        public const string Equals = "\uf52c";
        public const string Eraser = "\uf12d";
        public const string Ethernet = "\uf796";
        public const string EuroSign = "\uf153";
        public const string Exchange = "\uf0ec";
        public const string ExchangeAlt = "\uf362";
        public const string Exclamation = "\uf12a";
        public const string ExclamationCircle = "\uf06a";
        public const string ExclamationSquare = "\uf321";
        public const string ExclamationTriangle = "\uf071";
        public const string Expand = "\uf065";
        public const string ExpandAlt = "\uf424";
        public const string ExpandArrows = "\uf31d";
        public const string ExpandArrowsAlt = "\uf31e";
        public const string ExpandWide = "\uf320";
        public const string ExternalLink = "\uf08e";
        public const string ExternalLinkAlt = "\uf35d";
        public const string ExternalLinkSquare = "\uf14c";
        public const string ExternalLinkSquareAlt = "\uf360";
        public const string Eye = "\uf06e";
        public const string EyeDropper = "\uf1fb";
        public const string EyeEvil = "\uf6db";
        public const string EyeSlash = "\uf070";
        public const string Fan = "\uf863";
        public const string FanTable = "\ue004";
        public const string Farm = "\uf864";
        public const string FastBackward = "\uf049";
        public const string FastForward = "\uf050";
        public const string Faucet = "\ue005";
        public const string FaucetDrip = "\ue006";
        public const string Fax = "\uf1ac";
        public const string Feather = "\uf52d";
        public const string FeatherAlt = "\uf56b";
        public const string Female = "\uf182";
        public const string FieldHockey = "\uf44c";
        public const string FighterJet = "\uf0fb";
        public const string File = "\uf15b";
        public const string FileAlt = "\uf15c";
        public const string FileArchive = "\uf1c6";
        public const string FileAudio = "\uf1c7";
        public const string FileCertificate = "\uf5f3";
        public const string FileChartLine = "\uf659";
        public const string FileChartPie = "\uf65a";
        public const string FileCheck = "\uf316";
        public const string FileCode = "\uf1c9";
        public const string FileContract = "\uf56c";
        public const string FileCsv = "\uf6dd";
        public const string FileDownload = "\uf56d";
        public const string FileEdit = "\uf31c";
        public const string FileExcel = "\uf1c3";
        public const string FileExclamation = "\uf31a";
        public const string FileExport = "\uf56e";
        public const string FileImage = "\uf1c5";
        public const string FileImport = "\uf56f";
        public const string FileInvoice = "\uf570";
        public const string FileInvoiceDollar = "\uf571";
        public const string FileMedical = "\uf477";
        public const string FileMedicalAlt = "\uf478";
        public const string FileMinus = "\uf318";
        public const string FileMusic = "\uf8b6";
        public const string FilePdf = "\uf1c1";
        public const string FilePlus = "\uf319";
        public const string FilePowerpoint = "\uf1c4";
        public const string FilePrescription = "\uf572";
        public const string FileSearch = "\uf865";
        public const string FileSignature = "\uf573";
        public const string FileSpreadsheet = "\uf65b";
        public const string FileTimes = "\uf317";
        public const string FileUpload = "\uf574";
        public const string FileUser = "\uf65c";
        public const string FileVideo = "\uf1c8";
        public const string FileWord = "\uf1c2";
        public const string FilesMedical = "\uf7fd";
        public const string Fill = "\uf575";
        public const string FillDrip = "\uf576";
        public const string Film = "\uf008";
        public const string FilmAlt = "\uf3a0";
        public const string FilmCanister = "\uf8b7";
        public const string Filter = "\uf0b0";
        public const string Fingerprint = "\uf577";
        public const string Fire = "\uf06d";
        public const string FireAlt = "\uf7e4";
        public const string FireExtinguisher = "\uf134";
        public const string FireSmoke = "\uf74b";
        public const string Fireplace = "\uf79a";
        public const string FirstAid = "\uf479";
        public const string Fish = "\uf578";
        public const string FishCooked = "\uf7fe";
        public const string FistRaised = "\uf6de";
        public const string Flag = "\uf024";
        public const string FlagAlt = "\uf74c";
        public const string FlagCheckered = "\uf11e";
        public const string FlagUsa = "\uf74d";
        public const string Flame = "\uf6df";
        public const string Flashlight = "\uf8b8";
        public const string Flask = "\uf0c3";
        public const string FlaskPoison = "\uf6e0";
        public const string FlaskPotion = "\uf6e1";
        public const string Flower = "\uf7ff";
        public const string FlowerDaffodil = "\uf800";
        public const string FlowerTulip = "\uf801";
        public const string Flushed = "\uf579";
        public const string Flute = "\uf8b9";
        public const string FluxCapacitor = "\uf8ba";
        public const string Fog = "\uf74e";
        public const string Folder = "\uf07b";
        public const string FolderDownload = "\ue053";
        public const string FolderMinus = "\uf65d";
        public const string FolderOpen = "\uf07c";
        public const string FolderPlus = "\uf65e";
        public const string FolderTimes = "\uf65f";
        public const string FolderTree = "\uf802";
        public const string FolderUpload = "\ue054";
        public const string Folders = "\uf660";
        public const string Font = "\uf031";
        public const string FontAwesomeLogoFull = "\uf4e6";
        public const string FontCase = "\uf866";
        public const string FootballBall = "\uf44e";
        public const string FootballHelmet = "\uf44f";
        public const string Forklift = "\uf47a";
        public const string Forward = "\uf04e";
        public const string Fragile = "\uf4bb";
        public const string FrenchFries = "\uf803";
        public const string Frog = "\uf52e";
        public const string FrostyHead = "\uf79b";
        public const string Frown = "\uf119";
        public const string FrownOpen = "\uf57a";
        public const string Function = "\uf661";
        public const string FunnelDollar = "\uf662";
        public const string Futbol = "\uf1e3";
        public const string Galaxy = "\ue008";
        public const string GameBoard = "\uf867";
        public const string GameBoardAlt = "\uf868";
        public const string GameConsoleHandheld = "\uf8bb";
        public const string Gamepad = "\uf11b";
        public const string GamepadAlt = "\uf8bc";
        public const string Garage = "\ue009";
        public const string GarageCar = "\ue00a";
        public const string GarageOpen = "\ue00b";
        public const string GasPump = "\uf52f";
        public const string GasPumpSlash = "\uf5f4";
        public const string Gavel = "\uf0e3";
        public const string Gem = "\uf3a5";
        public const string Genderless = "\uf22d";
        public const string Ghost = "\uf6e2";
        public const string Gift = "\uf06b";
        public const string GiftCard = "\uf663";
        public const string Gifts = "\uf79c";
        public const string GingerbreadMan = "\uf79d";
        public const string Glass = "\uf804";
        public const string GlassChampagne = "\uf79e";
        public const string GlassCheers = "\uf79f";
        public const string GlassCitrus = "\uf869";
        public const string GlassMartini = "\uf000";
        public const string GlassMartiniAlt = "\uf57b";
        public const string GlassWhiskey = "\uf7a0";
        public const string GlassWhiskeyRocks = "\uf7a1";
        public const string Glasses = "\uf530";
        public const string GlassesAlt = "\uf5f5";
        public const string Globe = "\uf0ac";
        public const string GlobeAfrica = "\uf57c";
        public const string GlobeAmericas = "\uf57d";
        public const string GlobeAsia = "\uf57e";
        public const string GlobeEurope = "\uf7a2";
        public const string GlobeSnow = "\uf7a3";
        public const string GlobeStand = "\uf5f6";
        public const string GolfBall = "\uf450";
        public const string GolfClub = "\uf451";
        public const string Gopuram = "\uf664";
        public const string GraduationCap = "\uf19d";
        public const string Gramophone = "\uf8bd";
        public const string GreaterThan = "\uf531";
        public const string GreaterThanEqual = "\uf532";
        public const string Grimace = "\uf57f";
        public const string Grin = "\uf580";
        public const string GrinAlt = "\uf581";
        public const string GrinBeam = "\uf582";
        public const string GrinBeamSweat = "\uf583";
        public const string GrinHearts = "\uf584";
        public const string GrinSquint = "\uf585";
        public const string GrinSquintTears = "\uf586";
        public const string GrinStars = "\uf587";
        public const string GrinTears = "\uf588";
        public const string GrinTongue = "\uf589";
        public const string GrinTongueSquint = "\uf58a";
        public const string GrinTongueWink = "\uf58b";
        public const string GrinWink = "\uf58c";
        public const string GripHorizontal = "\uf58d";
        public const string GripLines = "\uf7a4";
        public const string GripLinesVertical = "\uf7a5";
        public const string GripVertical = "\uf58e";
        public const string Guitar = "\uf7a6";
        public const string GuitarElectric = "\uf8be";
        public const string Guitars = "\uf8bf";
        public const string HSquare = "\uf0fd";
        public const string H1 = "\uf313";
        public const string H2 = "\uf314";
        public const string H3 = "\uf315";
        public const string H4 = "\uf86a";
        public const string Hamburger = "\uf805";
        public const string Hammer = "\uf6e3";
        public const string HammerWar = "\uf6e4";
        public const string Hamsa = "\uf665";
        public const string HandHeart = "\uf4bc";
        public const string HandHolding = "\uf4bd";
        public const string HandHoldingBox = "\uf47b";
        public const string HandHoldingHeart = "\uf4be";
        public const string HandHoldingMagic = "\uf6e5";
        public const string HandHoldingMedical = "\ue05c";
        public const string HandHoldingSeedling = "\uf4bf";
        public const string HandHoldingUsd = "\uf4c0";
        public const string HandHoldingWater = "\uf4c1";
        public const string HandLizard = "\uf258";
        public const string HandMiddleFinger = "\uf806";
        public const string HandPaper = "\uf256";
        public const string HandPeace = "\uf25b";
        public const string HandPointDown = "\uf0a7";
        public const string HandPointLeft = "\uf0a5";
        public const string HandPointRight = "\uf0a4";
        public const string HandPointUp = "\uf0a6";
        public const string HandPointer = "\uf25a";
        public const string HandReceiving = "\uf47c";
        public const string HandRock = "\uf255";
        public const string HandScissors = "\uf257";
        public const string HandSparkles = "\ue05d";
        public const string HandSpock = "\uf259";
        public const string Hands = "\uf4c2";
        public const string HandsHeart = "\uf4c3";
        public const string HandsHelping = "\uf4c4";
        public const string HandsUsd = "\uf4c5";
        public const string HandsWash = "\ue05e";
        public const string Handshake = "\uf2b5";
        public const string HandshakeAlt = "\uf4c6";
        public const string HandshakeAltSlash = "\ue05f";
        public const string HandshakeSlash = "\ue060";
        public const string Hanukiah = "\uf6e6";
        public const string HardHat = "\uf807";
        public const string Hashtag = "\uf292";
        public const string HatChef = "\uf86b";
        public const string HatCowboy = "\uf8c0";
        public const string HatCowboySide = "\uf8c1";
        public const string HatSanta = "\uf7a7";
        public const string HatWinter = "\uf7a8";
        public const string HatWitch = "\uf6e7";
        public const string HatWizard = "\uf6e8";
        public const string Hdd = "\uf0a0";
        public const string HeadSide = "\uf6e9";
        public const string HeadSideBrain = "\uf808";
        public const string HeadSideCough = "\ue061";
        public const string HeadSideCoughSlash = "\ue062";
        public const string HeadSideHeadphones = "\uf8c2";
        public const string HeadSideMask = "\ue063";
        public const string HeadSideMedical = "\uf809";
        public const string HeadSideVirus = "\ue064";
        public const string HeadVr = "\uf6ea";
        public const string Heading = "\uf1dc";
        public const string Headphones = "\uf025";
        public const string HeadphonesAlt = "\uf58f";
        public const string Headset = "\uf590";
        public const string Heart = "\uf004";
        public const string HeartBroken = "\uf7a9";
        public const string HeartCircle = "\uf4c7";
        public const string HeartRate = "\uf5f8";
        public const string HeartSquare = "\uf4c8";
        public const string Heartbeat = "\uf21e";
        public const string Heat = "\ue00c";
        public const string Helicopter = "\uf533";
        public const string HelmetBattle = "\uf6eb";
        public const string Hexagon = "\uf312";
        public const string Highlighter = "\uf591";
        public const string Hiking = "\uf6ec";
        public const string Hippo = "\uf6ed";
        public const string History = "\uf1da";
        public const string HockeyMask = "\uf6ee";
        public const string HockeyPuck = "\uf453";
        public const string HockeySticks = "\uf454";
        public const string HollyBerry = "\uf7aa";
        public const string Home = "\uf015";
        public const string HomeAlt = "\uf80a";
        public const string HomeHeart = "\uf4c9";
        public const string HomeLg = "\uf80b";
        public const string HomeLgAlt = "\uf80c";
        public const string HoodCloak = "\uf6ef";
        public const string HorizontalRule = "\uf86c";
        public const string Horse = "\uf6f0";
        public const string HorseHead = "\uf7ab";
        public const string HorseSaddle = "\uf8c3";
        public const string Hospital = "\uf0f8";
        public const string HospitalAlt = "\uf47d";
        public const string HospitalSymbol = "\uf47e";
        public const string HospitalUser = "\uf80d";
        public const string Hospitals = "\uf80e";
        public const string HotTub = "\uf593";
        public const string Hotdog = "\uf80f";
        public const string Hotel = "\uf594";
        public const string Hourglass = "\uf254";
        public const string HourglassEnd = "\uf253";
        public const string HourglassHalf = "\uf252";
        public const string HourglassStart = "\uf251";
        public const string House = "\ue00d";
        public const string HouseDamage = "\uf6f1";
        public const string HouseDay = "\ue00e";
        public const string HouseFlood = "\uf74f";
        public const string HouseLeave = "\ue00f";
        public const string HouseNight = "\ue010";
        public const string HouseReturn = "\ue011";
        public const string HouseSignal = "\ue012";
        public const string HouseUser = "\ue065";
        public const string Hryvnia = "\uf6f2";
        public const string Humidity = "\uf750";
        public const string Hurricane = "\uf751";
        public const string ICursor = "\uf246";
        public const string IceCream = "\uf810";
        public const string IceSkate = "\uf7ac";
        public const string Icicles = "\uf7ad";
        public const string Icons = "\uf86d";
        public const string IconsAlt = "\uf86e";
        public const string IdBadge = "\uf2c1";
        public const string IdCard = "\uf2c2";
        public const string IdCardAlt = "\uf47f";
        public const string Igloo = "\uf7ae";
        public const string Image = "\uf03e";
        public const string ImagePolaroid = "\uf8c4";
        public const string Images = "\uf302";
        public const string Inbox = "\uf01c";
        public const string InboxIn = "\uf310";
        public const string InboxOut = "\uf311";
        public const string Indent = "\uf03c";
        public const string Industry = "\uf275";
        public const string IndustryAlt = "\uf3b3";
        public const string Infinity = "\uf534";
        public const string Info = "\uf129";
        public const string InfoCircle = "\uf05a";
        public const string InfoSquare = "\uf30f";
        public const string Inhaler = "\uf5f9";
        public const string Integral = "\uf667";
        public const string Intersection = "\uf668";
        public const string Inventory = "\uf480";
        public const string IslandTropical = "\uf811";
        public const string Italic = "\uf033";
        public const string JackOLantern = "\uf30e";
        public const string Jedi = "\uf669";
        public const string Joint = "\uf595";
        public const string JournalWhills = "\uf66a";
        public const string Joystick = "\uf8c5";
        public const string Jug = "\uf8c6";
        public const string Kaaba = "\uf66b";
        public const string Kazoo = "\uf8c7";
        public const string Kerning = "\uf86f";
        public const string Key = "\uf084";
        public const string KeySkeleton = "\uf6f3";
        public const string Keyboard = "\uf11c";
        public const string Keynote = "\uf66c";
        public const string Khanda = "\uf66d";
        public const string Kidneys = "\uf5fb";
        public const string Kiss = "\uf596";
        public const string KissBeam = "\uf597";
        public const string KissWinkHeart = "\uf598";
        public const string Kite = "\uf6f4";
        public const string KiwiBird = "\uf535";
        public const string KnifeKitchen = "\uf6f5";
        public const string Lambda = "\uf66e";
        public const string Lamp = "\uf4ca";
        public const string LampDesk = "\ue014";
        public const string LampFloor = "\ue015";
        public const string Landmark = "\uf66f";
        public const string LandmarkAlt = "\uf752";
        public const string Language = "\uf1ab";
        public const string Laptop = "\uf109";
        public const string LaptopCode = "\uf5fc";
        public const string LaptopHouse = "\ue066";
        public const string LaptopMedical = "\uf812";
        public const string Lasso = "\uf8c8";
        public const string Laugh = "\uf599";
        public const string LaughBeam = "\uf59a";
        public const string LaughSquint = "\uf59b";
        public const string LaughWink = "\uf59c";
        public const string LayerGroup = "\uf5fd";
        public const string LayerMinus = "\uf5fe";
        public const string LayerPlus = "\uf5ff";
        public const string Leaf = "\uf06c";
        public const string LeafHeart = "\uf4cb";
        public const string LeafMaple = "\uf6f6";
        public const string LeafOak = "\uf6f7";
        public const string Lemon = "\uf094";
        public const string LessThan = "\uf536";
        public const string LessThanEqual = "\uf537";
        public const string LevelDown = "\uf149";
        public const string LevelDownAlt = "\uf3be";
        public const string LevelUp = "\uf148";
        public const string LevelUpAlt = "\uf3bf";
        public const string LifeRing = "\uf1cd";
        public const string LightCeiling = "\ue016";
        public const string LightSwitch = "\ue017";
        public const string LightSwitchOff = "\ue018";
        public const string LightSwitchOn = "\ue019";
        public const string Lightbulb = "\uf0eb";
        public const string LightbulbDollar = "\uf670";
        public const string LightbulbExclamation = "\uf671";
        public const string LightbulbOn = "\uf672";
        public const string LightbulbSlash = "\uf673";
        public const string LightsHoliday = "\uf7b2";
        public const string LineColumns = "\uf870";
        public const string LineHeight = "\uf871";
        public const string Link = "\uf0c1";
        public const string Lips = "\uf600";
        public const string LiraSign = "\uf195";
        public const string List = "\uf03a";
        public const string ListAlt = "\uf022";
        public const string ListMusic = "\uf8c9";
        public const string ListOl = "\uf0cb";
        public const string ListUl = "\uf0ca";
        public const string Location = "\uf601";
        public const string LocationArrow = "\uf124";
        public const string LocationCircle = "\uf602";
        public const string LocationSlash = "\uf603";
        public const string Lock = "\uf023";
        public const string LockAlt = "\uf30d";
        public const string LockOpen = "\uf3c1";
        public const string LockOpenAlt = "\uf3c2";
        public const string LongArrowAltDown = "\uf309";
        public const string LongArrowAltLeft = "\uf30a";
        public const string LongArrowAltRight = "\uf30b";
        public const string LongArrowAltUp = "\uf30c";
        public const string LongArrowDown = "\uf175";
        public const string LongArrowLeft = "\uf177";
        public const string LongArrowRight = "\uf178";
        public const string LongArrowUp = "\uf176";
        public const string Loveseat = "\uf4cc";
        public const string LowVision = "\uf2a8";
        public const string Luchador = "\uf455";
        public const string LuggageCart = "\uf59d";
        public const string Lungs = "\uf604";
        public const string LungsVirus = "\ue067";
        public const string Mace = "\uf6f8";
        public const string Magic = "\uf0d0";
        public const string Magnet = "\uf076";
        public const string MailBulk = "\uf674";
        public const string Mailbox = "\uf813";
        public const string Male = "\uf183";
        public const string Mandolin = "\uf6f9";
        public const string Map = "\uf279";
        public const string MapMarked = "\uf59f";
        public const string MapMarkedAlt = "\uf5a0";
        public const string MapMarker = "\uf041";
        public const string MapMarkerAlt = "\uf3c5";
        public const string MapMarkerAltSlash = "\uf605";
        public const string MapMarkerCheck = "\uf606";
        public const string MapMarkerEdit = "\uf607";
        public const string MapMarkerExclamation = "\uf608";
        public const string MapMarkerMinus = "\uf609";
        public const string MapMarkerPlus = "\uf60a";
        public const string MapMarkerQuestion = "\uf60b";
        public const string MapMarkerSlash = "\uf60c";
        public const string MapMarkerSmile = "\uf60d";
        public const string MapMarkerTimes = "\uf60e";
        public const string MapPin = "\uf276";
        public const string MapSigns = "\uf277";
        public const string Marker = "\uf5a1";
        public const string Mars = "\uf222";
        public const string MarsDouble = "\uf227";
        public const string MarsStroke = "\uf229";
        public const string MarsStrokeH = "\uf22b";
        public const string MarsStrokeV = "\uf22a";
        public const string Mask = "\uf6fa";
        public const string Meat = "\uf814";
        public const string Medal = "\uf5a2";
        public const string Medkit = "\uf0fa";
        public const string Megaphone = "\uf675";
        public const string Meh = "\uf11a";
        public const string MehBlank = "\uf5a4";
        public const string MehRollingEyes = "\uf5a5";
        public const string Memory = "\uf538";
        public const string Menorah = "\uf676";
        public const string Mercury = "\uf223";
        public const string Meteor = "\uf753";
        public const string Microchip = "\uf2db";
        public const string Microphone = "\uf130";
        public const string MicrophoneAlt = "\uf3c9";
        public const string MicrophoneAltSlash = "\uf539";
        public const string MicrophoneSlash = "\uf131";
        public const string MicrophoneStand = "\uf8cb";
        public const string Microscope = "\uf610";
        public const string Microwave = "\ue01b";
        public const string MindShare = "\uf677";
        public const string Minus = "\uf068";
        public const string MinusCircle = "\uf056";
        public const string MinusHexagon = "\uf307";
        public const string MinusOctagon = "\uf308";
        public const string MinusSquare = "\uf146";
        public const string Mistletoe = "\uf7b4";
        public const string Mitten = "\uf7b5";
        public const string Mobile = "\uf10b";
        public const string MobileAlt = "\uf3cd";
        public const string MobileAndroid = "\uf3ce";
        public const string MobileAndroidAlt = "\uf3cf";
        public const string MoneyBill = "\uf0d6";
        public const string MoneyBillAlt = "\uf3d1";
        public const string MoneyBillWave = "\uf53a";
        public const string MoneyBillWaveAlt = "\uf53b";
        public const string MoneyCheck = "\uf53c";
        public const string MoneyCheckAlt = "\uf53d";
        public const string MoneyCheckEdit = "\uf872";
        public const string MoneyCheckEditAlt = "\uf873";
        public const string MonitorHeartRate = "\uf611";
        public const string Monkey = "\uf6fb";
        public const string Monument = "\uf5a6";
        public const string Moon = "\uf186";
        public const string MoonCloud = "\uf754";
        public const string MoonStars = "\uf755";
        public const string MortarPestle = "\uf5a7";
        public const string Mosque = "\uf678";
        public const string Motorcycle = "\uf21c";
        public const string Mountain = "\uf6fc";
        public const string Mountains = "\uf6fd";
        public const string Mouse = "\uf8cc";
        public const string MouseAlt = "\uf8cd";
        public const string MousePointer = "\uf245";
        public const string Mp3Player = "\uf8ce";
        public const string Mug = "\uf874";
        public const string MugHot = "\uf7b6";
        public const string MugMarshmallows = "\uf7b7";
        public const string MugTea = "\uf875";
        public const string Music = "\uf001";
        public const string MusicAlt = "\uf8cf";
        public const string MusicAltSlash = "\uf8d0";
        public const string MusicSlash = "\uf8d1";
        public const string Narwhal = "\uf6fe";
        public const string NetworkWired = "\uf6ff";
        public const string Neuter = "\uf22c";
        public const string Newspaper = "\uf1ea";
        public const string NotEqual = "\uf53e";
        public const string NotesMedical = "\uf481";
        public const string ObjectGroup = "\uf247";
        public const string ObjectUngroup = "\uf248";
        public const string Octagon = "\uf306";
        public const string OilCan = "\uf613";
        public const string OilTemp = "\uf614";
        public const string Om = "\uf679";
        public const string Omega = "\uf67a";
        public const string Ornament = "\uf7b8";
        public const string Otter = "\uf700";
        public const string Outdent = "\uf03b";
        public const string Outlet = "\ue01c";
        public const string Oven = "\ue01d";
        public const string Overline = "\uf876";
        public const string PageBreak = "\uf877";
        public const string Pager = "\uf815";
        public const string PaintBrush = "\uf1fc";
        public const string PaintBrushAlt = "\uf5a9";
        public const string PaintRoller = "\uf5aa";
        public const string Palette = "\uf53f";
        public const string Pallet = "\uf482";
        public const string PalletAlt = "\uf483";
        public const string PaperPlane = "\uf1d8";
        public const string Paperclip = "\uf0c6";
        public const string ParachuteBox = "\uf4cd";
        public const string Paragraph = "\uf1dd";
        public const string ParagraphRtl = "\uf878";
        public const string Parking = "\uf540";
        public const string ParkingCircle = "\uf615";
        public const string ParkingCircleSlash = "\uf616";
        public const string ParkingSlash = "\uf617";
        public const string Passport = "\uf5ab";
        public const string Pastafarianism = "\uf67b";
        public const string Paste = "\uf0ea";
        public const string Pause = "\uf04c";
        public const string PauseCircle = "\uf28b";
        public const string Paw = "\uf1b0";
        public const string PawAlt = "\uf701";
        public const string PawClaws = "\uf702";
        public const string Peace = "\uf67c";
        public const string Pegasus = "\uf703";
        public const string Pen = "\uf304";
        public const string PenAlt = "\uf305";
        public const string PenFancy = "\uf5ac";
        public const string PenNib = "\uf5ad";
        public const string PenSquare = "\uf14b";
        public const string Pencil = "\uf040";
        public const string PencilAlt = "\uf303";
        public const string PencilPaintbrush = "\uf618";
        public const string PencilRuler = "\uf5ae";
        public const string Pennant = "\uf456";
        public const string PeopleArrows = "\ue068";
        public const string PeopleCarry = "\uf4ce";
        public const string PepperHot = "\uf816";
        public const string Percent = "\uf295";
        public const string Percentage = "\uf541";
        public const string PersonBooth = "\uf756";
        public const string PersonCarry = "\uf4cf";
        public const string PersonDolly = "\uf4d0";
        public const string PersonDollyEmpty = "\uf4d1";
        public const string PersonSign = "\uf757";
        public const string Phone = "\uf095";
        public const string PhoneAlt = "\uf879";
        public const string PhoneLaptop = "\uf87a";
        public const string PhoneOffice = "\uf67d";
        public const string PhonePlus = "\uf4d2";
        public const string PhoneRotary = "\uf8d3";
        public const string PhoneSlash = "\uf3dd";
        public const string PhoneSquare = "\uf098";
        public const string PhoneSquareAlt = "\uf87b";
        public const string PhoneVolume = "\uf2a0";
        public const string PhotoVideo = "\uf87c";
        public const string Pi = "\uf67e";
        public const string Piano = "\uf8d4";
        public const string PianoKeyboard = "\uf8d5";
        public const string Pie = "\uf705";
        public const string Pig = "\uf706";
        public const string PiggyBank = "\uf4d3";
        public const string Pills = "\uf484";
        public const string Pizza = "\uf817";
        public const string PizzaSlice = "\uf818";
        public const string PlaceOfWorship = "\uf67f";
        public const string Plane = "\uf072";
        public const string PlaneAlt = "\uf3de";
        public const string PlaneArrival = "\uf5af";
        public const string PlaneDeparture = "\uf5b0";
        public const string PlaneSlash = "\ue069";
        public const string PlanetMoon = "\ue01f";
        public const string PlanetRinged = "\ue020";
        public const string Play = "\uf04b";
        public const string PlayCircle = "\uf144";
        public const string Plug = "\uf1e6";
        public const string Plus = "\uf067";
        public const string PlusCircle = "\uf055";
        public const string PlusHexagon = "\uf300";
        public const string PlusOctagon = "\uf301";
        public const string PlusSquare = "\uf0fe";
        public const string Podcast = "\uf2ce";
        public const string Podium = "\uf680";
        public const string PodiumStar = "\uf758";
        public const string PoliceBox = "\ue021";
        public const string Poll = "\uf681";
        public const string PollH = "\uf682";
        public const string PollPeople = "\uf759";
        public const string Poo = "\uf2fe";
        public const string PooStorm = "\uf75a";
        public const string Poop = "\uf619";
        public const string Popcorn = "\uf819";
        public const string PortalEnter = "\ue022";
        public const string PortalExit = "\ue023";
        public const string Portrait = "\uf3e0";
        public const string PoundSign = "\uf154";
        public const string PowerOff = "\uf011";
        public const string Pray = "\uf683";
        public const string PrayingHands = "\uf684";
        public const string Prescription = "\uf5b1";
        public const string PrescriptionBottle = "\uf485";
        public const string PrescriptionBottleAlt = "\uf486";
        public const string Presentation = "\uf685";
        public const string Print = "\uf02f";
        public const string PrintSearch = "\uf81a";
        public const string PrintSlash = "\uf686";
        public const string Procedures = "\uf487";
        public const string ProjectDiagram = "\uf542";
        public const string Projector = "\uf8d6";
        public const string PumpMedical = "\ue06a";
        public const string PumpSoap = "\ue06b";
        public const string Pumpkin = "\uf707";
        public const string PuzzlePiece = "\uf12e";
        public const string Qrcode = "\uf029";
        public const string Question = "\uf128";
        public const string QuestionCircle = "\uf059";
        public const string QuestionSquare = "\uf2fd";
        public const string Quidditch = "\uf458";
        public const string QuoteLeft = "\uf10d";
        public const string QuoteRight = "\uf10e";
        public const string Quran = "\uf687";
        public const string Rabbit = "\uf708";
        public const string RabbitFast = "\uf709";
        public const string Racquet = "\uf45a";
        public const string Radar = "\ue024";
        public const string Radiation = "\uf7b9";
        public const string RadiationAlt = "\uf7ba";
        public const string Radio = "\uf8d7";
        public const string RadioAlt = "\uf8d8";
        public const string Rainbow = "\uf75b";
        public const string Raindrops = "\uf75c";
        public const string Ram = "\uf70a";
        public const string RampLoading = "\uf4d4";
        public const string Random = "\uf074";
        public const string Raygun = "\ue025";
        public const string Receipt = "\uf543";
        public const string RecordVinyl = "\uf8d9";
        public const string RectangleLandscape = "\uf2fa";
        public const string RectanglePortrait = "\uf2fb";
        public const string RectangleWide = "\uf2fc";
        public const string Recycle = "\uf1b8";
        public const string Redo = "\uf01e";
        public const string RedoAlt = "\uf2f9";
        public const string Refrigerator = "\ue026";
        public const string Registered = "\uf25d";
        public const string RemoveFormat = "\uf87d";
        public const string Repeat = "\uf363";
        public const string Repeat1 = "\uf365";
        public const string Repeat1Alt = "\uf366";
        public const string RepeatAlt = "\uf364";
        public const string Reply = "\uf3e5";
        public const string ReplyAll = "\uf122";
        public const string Republican = "\uf75e";
        public const string Restroom = "\uf7bd";
        public const string Retweet = "\uf079";
        public const string RetweetAlt = "\uf361";
        public const string Ribbon = "\uf4d6";
        public const string Ring = "\uf70b";
        public const string RingsWedding = "\uf81b";
        public const string Road = "\uf018";
        public const string Robot = "\uf544";
        public const string Rocket = "\uf135";
        public const string RocketLaunch = "\ue027";
        public const string Route = "\uf4d7";
        public const string RouteHighway = "\uf61a";
        public const string RouteInterstate = "\uf61b";
        public const string Router = "\uf8da";
        public const string Rss = "\uf09e";
        public const string RssSquare = "\uf143";
        public const string RubleSign = "\uf158";
        public const string Ruler = "\uf545";
        public const string RulerCombined = "\uf546";
        public const string RulerHorizontal = "\uf547";
        public const string RulerTriangle = "\uf61c";
        public const string RulerVertical = "\uf548";
        public const string Running = "\uf70c";
        public const string RupeeSign = "\uf156";
        public const string Rv = "\uf7be";
        public const string Sack = "\uf81c";
        public const string SackDollar = "\uf81d";
        public const string SadCry = "\uf5b3";
        public const string SadTear = "\uf5b4";
        public const string Salad = "\uf81e";
        public const string Sandwich = "\uf81f";
        public const string Satellite = "\uf7bf";
        public const string SatelliteDish = "\uf7c0";
        public const string Sausage = "\uf820";
        public const string Save = "\uf0c7";
        public const string SaxHot = "\uf8db";
        public const string Saxophone = "\uf8dc";
        public const string Scalpel = "\uf61d";
        public const string ScalpelPath = "\uf61e";
        public const string Scanner = "\uf488";
        public const string ScannerImage = "\uf8f3";
        public const string ScannerKeyboard = "\uf489";
        public const string ScannerTouchscreen = "\uf48a";
        public const string Scarecrow = "\uf70d";
        public const string Scarf = "\uf7c1";
        public const string School = "\uf549";
        public const string Screwdriver = "\uf54a";
        public const string Scroll = "\uf70e";
        public const string ScrollOld = "\uf70f";
        public const string Scrubber = "\uf2f8";
        public const string Scythe = "\uf710";
        public const string SdCard = "\uf7c2";
        public const string Search = "\uf002";
        public const string SearchDollar = "\uf688";
        public const string SearchLocation = "\uf689";
        public const string SearchMinus = "\uf010";
        public const string SearchPlus = "\uf00e";
        public const string Seedling = "\uf4d8";
        public const string SendBack = "\uf87e";
        public const string SendBackward = "\uf87f";
        public const string Sensor = "\ue028";
        public const string SensorAlert = "\ue029";
        public const string SensorFire = "\ue02a";
        public const string SensorOn = "\ue02b";
        public const string SensorSmoke = "\ue02c";
        public const string Server = "\uf233";
        public const string Shapes = "\uf61f";
        public const string Share = "\uf064";
        public const string ShareAll = "\uf367";
        public const string ShareAlt = "\uf1e0";
        public const string ShareAltSquare = "\uf1e1";
        public const string ShareSquare = "\uf14d";
        public const string Sheep = "\uf711";
        public const string ShekelSign = "\uf20b";
        public const string Shield = "\uf132";
        public const string ShieldAlt = "\uf3ed";
        public const string ShieldCheck = "\uf2f7";
        public const string ShieldCross = "\uf712";
        public const string ShieldVirus = "\ue06c";
        public const string Ship = "\uf21a";
        public const string ShippingFast = "\uf48b";
        public const string ShippingTimed = "\uf48c";
        public const string ShishKebab = "\uf821";
        public const string ShoePrints = "\uf54b";
        public const string ShoppingBag = "\uf290";
        public const string ShoppingBasket = "\uf291";
        public const string ShoppingCart = "\uf07a";
        public const string Shovel = "\uf713";
        public const string ShovelSnow = "\uf7c3";
        public const string Shower = "\uf2cc";
        public const string Shredder = "\uf68a";
        public const string ShuttleVan = "\uf5b6";
        public const string Shuttlecock = "\uf45b";
        public const string Sickle = "\uf822";
        public const string Sigma = "\uf68b";
        public const string Sign = "\uf4d9";
        public const string SignIn = "\uf090";
        public const string SignInAlt = "\uf2f6";
        public const string SignLanguage = "\uf2a7";
        public const string SignOut = "\uf08b";
        public const string SignOutAlt = "\uf2f5";
        public const string Signal = "\uf012";
        public const string Signal1 = "\uf68c";
        public const string Signal2 = "\uf68d";
        public const string Signal3 = "\uf68e";
        public const string Signal4 = "\uf68f";
        public const string SignalAlt = "\uf690";
        public const string SignalAlt1 = "\uf691";
        public const string SignalAlt2 = "\uf692";
        public const string SignalAlt3 = "\uf693";
        public const string SignalAltSlash = "\uf694";
        public const string SignalSlash = "\uf695";
        public const string SignalStream = "\uf8dd";
        public const string Signature = "\uf5b7";
        public const string SimCard = "\uf7c4";
        public const string Sink = "\ue06d";
        public const string Siren = "\ue02d";
        public const string SirenOn = "\ue02e";
        public const string Sitemap = "\uf0e8";
        public const string Skating = "\uf7c5";
        public const string Skeleton = "\uf620";
        public const string SkiJump = "\uf7c7";
        public const string SkiLift = "\uf7c8";
        public const string Skiing = "\uf7c9";
        public const string SkiingNordic = "\uf7ca";
        public const string Skull = "\uf54c";
        public const string SkullCow = "\uf8de";
        public const string SkullCrossbones = "\uf714";
        public const string Slash = "\uf715";
        public const string Sledding = "\uf7cb";
        public const string Sleigh = "\uf7cc";
        public const string SlidersH = "\uf1de";
        public const string SlidersHSquare = "\uf3f0";
        public const string SlidersV = "\uf3f1";
        public const string SlidersVSquare = "\uf3f2";
        public const string Smile = "\uf118";
        public const string SmileBeam = "\uf5b8";
        public const string SmilePlus = "\uf5b9";
        public const string SmileWink = "\uf4da";
        public const string Smog = "\uf75f";
        public const string Smoke = "\uf760";
        public const string Smoking = "\uf48d";
        public const string SmokingBan = "\uf54d";
        public const string Sms = "\uf7cd";
        public const string Snake = "\uf716";
        public const string Snooze = "\uf880";
        public const string SnowBlowing = "\uf761";
        public const string Snowboarding = "\uf7ce";
        public const string Snowflake = "\uf2dc";
        public const string Snowflakes = "\uf7cf";
        public const string Snowman = "\uf7d0";
        public const string Snowmobile = "\uf7d1";
        public const string Snowplow = "\uf7d2";
        public const string Soap = "\ue06e";
        public const string Socks = "\uf696";
        public const string SolarPanel = "\uf5ba";
        public const string SolarSystem = "\ue02f";
        public const string Sort = "\uf0dc";
        public const string SortAlphaDown = "\uf15d";
        public const string SortAlphaDownAlt = "\uf881";
        public const string SortAlphaUp = "\uf15e";
        public const string SortAlphaUpAlt = "\uf882";
        public const string SortAlt = "\uf883";
        public const string SortAmountDown = "\uf160";
        public const string SortAmountDownAlt = "\uf884";
        public const string SortAmountUp = "\uf161";
        public const string SortAmountUpAlt = "\uf885";
        public const string SortCircle = "\ue030";
        public const string SortCircleDown = "\ue031";
        public const string SortCircleUp = "\ue032";
        public const string SortDown = "\uf0dd";
        public const string SortNumericDown = "\uf162";
        public const string SortNumericDownAlt = "\uf886";
        public const string SortNumericUp = "\uf163";
        public const string SortNumericUpAlt = "\uf887";
        public const string SortShapesDown = "\uf888";
        public const string SortShapesDownAlt = "\uf889";
        public const string SortShapesUp = "\uf88a";
        public const string SortShapesUpAlt = "\uf88b";
        public const string SortSizeDown = "\uf88c";
        public const string SortSizeDownAlt = "\uf88d";
        public const string SortSizeUp = "\uf88e";
        public const string SortSizeUpAlt = "\uf88f";
        public const string SortUp = "\uf0de";
        public const string Soup = "\uf823";
        public const string Spa = "\uf5bb";
        public const string SpaceShuttle = "\uf197";
        public const string SpaceStationMoon = "\ue033";
        public const string SpaceStationMoonAlt = "\ue034";
        public const string Spade = "\uf2f4";
        public const string Sparkles = "\uf890";
        public const string Speaker = "\uf8df";
        public const string Speakers = "\uf8e0";
        public const string SpellCheck = "\uf891";
        public const string Spider = "\uf717";
        public const string SpiderBlackWidow = "\uf718";
        public const string SpiderWeb = "\uf719";
        public const string Spinner = "\uf110";
        public const string SpinnerThird = "\uf3f4";
        public const string Splotch = "\uf5bc";
        public const string SprayCan = "\uf5bd";
        public const string Sprinkler = "\ue035";
        public const string Square = "\uf0c8";
        public const string SquareFull = "\uf45c";
        public const string SquareRoot = "\uf697";
        public const string SquareRootAlt = "\uf698";
        public const string Squirrel = "\uf71a";
        public const string Staff = "\uf71b";
        public const string Stamp = "\uf5bf";
        public const string Star = "\uf005";
        public const string StarAndCrescent = "\uf699";
        public const string StarChristmas = "\uf7d4";
        public const string StarExclamation = "\uf2f3";
        public const string StarHalf = "\uf089";
        public const string StarHalfAlt = "\uf5c0";
        public const string StarOfDavid = "\uf69a";
        public const string StarOfLife = "\uf621";
        public const string StarShooting = "\ue036";
        public const string Starfighter = "\ue037";
        public const string StarfighterAlt = "\ue038";
        public const string Stars = "\uf762";
        public const string Starship = "\ue039";
        public const string StarshipFreighter = "\ue03a";
        public const string Steak = "\uf824";
        public const string SteeringWheel = "\uf622";
        public const string StepBackward = "\uf048";
        public const string StepForward = "\uf051";
        public const string Stethoscope = "\uf0f1";
        public const string StickyNote = "\uf249";
        public const string Stocking = "\uf7d5";
        public const string Stomach = "\uf623";
        public const string Stop = "\uf04d";
        public const string StopCircle = "\uf28d";
        public const string Stopwatch = "\uf2f2";
        public const string Stopwatch20 = "\ue06f";
        public const string Store = "\uf54e";
        public const string StoreAlt = "\uf54f";
        public const string StoreAltSlash = "\ue070";
        public const string StoreSlash = "\ue071";
        public const string Stream = "\uf550";
        public const string StreetView = "\uf21d";
        public const string Stretcher = "\uf825";
        public const string Strikethrough = "\uf0cc";
        public const string Stroopwafel = "\uf551";
        public const string Subscript = "\uf12c";
        public const string Subway = "\uf239";
        public const string Suitcase = "\uf0f2";
        public const string SuitcaseRolling = "\uf5c1";
        public const string Sun = "\uf185";
        public const string SunCloud = "\uf763";
        public const string SunDust = "\uf764";
        public const string SunHaze = "\uf765";
        public const string Sunglasses = "\uf892";
        public const string Sunrise = "\uf766";
        public const string Sunset = "\uf767";
        public const string Superscript = "\uf12b";
        public const string Surprise = "\uf5c2";
        public const string Swatchbook = "\uf5c3";
        public const string Swimmer = "\uf5c4";
        public const string SwimmingPool = "\uf5c5";
        public const string Sword = "\uf71c";
        public const string SwordLaser = "\ue03b";
        public const string SwordLaserAlt = "\ue03c";
        public const string Swords = "\uf71d";
        public const string SwordsLaser = "\ue03d";
        public const string Synagogue = "\uf69b";
        public const string Sync = "\uf021";
        public const string SyncAlt = "\uf2f1";
        public const string Syringe = "\uf48e";
        public const string Table = "\uf0ce";
        public const string TableTennis = "\uf45d";
        public const string Tablet = "\uf10a";
        public const string TabletAlt = "\uf3fa";
        public const string TabletAndroid = "\uf3fb";
        public const string TabletAndroidAlt = "\uf3fc";
        public const string TabletRugged = "\uf48f";
        public const string Tablets = "\uf490";
        public const string Tachometer = "\uf0e4";
        public const string TachometerAlt = "\uf3fd";
        public const string TachometerAltAverage = "\uf624";
        public const string TachometerAltFast = "\uf625";
        public const string TachometerAltFastest = "\uf626";
        public const string TachometerAltSlow = "\uf627";
        public const string TachometerAltSlowest = "\uf628";
        public const string TachometerAverage = "\uf629";
        public const string TachometerFast = "\uf62a";
        public const string TachometerFastest = "\uf62b";
        public const string TachometerSlow = "\uf62c";
        public const string TachometerSlowest = "\uf62d";
        public const string Taco = "\uf826";
        public const string Tag = "\uf02b";
        public const string Tags = "\uf02c";
        public const string Tally = "\uf69c";
        public const string Tanakh = "\uf827";
        public const string Tape = "\uf4db";
        public const string Tasks = "\uf0ae";
        public const string TasksAlt = "\uf828";
        public const string Taxi = "\uf1ba";
        public const string Teeth = "\uf62e";
        public const string TeethOpen = "\uf62f";
        public const string Telescope = "\ue03e";
        public const string TemperatureDown = "\ue03f";
        public const string TemperatureFrigid = "\uf768";
        public const string TemperatureHigh = "\uf769";
        public const string TemperatureHot = "\uf76a";
        public const string TemperatureLow = "\uf76b";
        public const string TemperatureUp = "\ue040";
        public const string Tenge = "\uf7d7";
        public const string TennisBall = "\uf45e";
        public const string Terminal = "\uf120";
        public const string Text = "\uf893";
        public const string TextHeight = "\uf034";
        public const string TextSize = "\uf894";
        public const string TextWidth = "\uf035";
        public const string Th = "\uf00a";
        public const string ThLarge = "\uf009";
        public const string ThList = "\uf00b";
        public const string TheaterMasks = "\uf630";
        public const string Thermometer = "\uf491";
        public const string ThermometerEmpty = "\uf2cb";
        public const string ThermometerFull = "\uf2c7";
        public const string ThermometerHalf = "\uf2c9";
        public const string ThermometerQuarter = "\uf2ca";
        public const string ThermometerThreeQuarters = "\uf2c8";
        public const string Theta = "\uf69e";
        public const string ThumbsDown = "\uf165";
        public const string ThumbsUp = "\uf164";
        public const string Thumbtack = "\uf08d";
        public const string Thunderstorm = "\uf76c";
        public const string ThunderstormMoon = "\uf76d";
        public const string ThunderstormSun = "\uf76e";
        public const string Ticket = "\uf145";
        public const string TicketAlt = "\uf3ff";
        public const string Tilde = "\uf69f";
        public const string Times = "\uf00d";
        public const string TimesCircle = "\uf057";
        public const string TimesHexagon = "\uf2ee";
        public const string TimesOctagon = "\uf2f0";
        public const string TimesSquare = "\uf2d3";
        public const string Tint = "\uf043";
        public const string TintSlash = "\uf5c7";
        public const string Tire = "\uf631";
        public const string TireFlat = "\uf632";
        public const string TirePressureWarning = "\uf633";
        public const string TireRugged = "\uf634";
        public const string Tired = "\uf5c8";
        public const string ToggleOff = "\uf204";
        public const string ToggleOn = "\uf205";
        public const string Toilet = "\uf7d8";
        public const string ToiletPaper = "\uf71e";
        public const string ToiletPaperAlt = "\uf71f";
        public const string ToiletPaperSlash = "\ue072";
        public const string Tombstone = "\uf720";
        public const string TombstoneAlt = "\uf721";
        public const string Toolbox = "\uf552";
        public const string Tools = "\uf7d9";
        public const string Tooth = "\uf5c9";
        public const string Toothbrush = "\uf635";
        public const string Torah = "\uf6a0";
        public const string ToriiGate = "\uf6a1";
        public const string Tornado = "\uf76f";
        public const string Tractor = "\uf722";
        public const string Trademark = "\uf25c";
        public const string TrafficCone = "\uf636";
        public const string TrafficLight = "\uf637";
        public const string TrafficLightGo = "\uf638";
        public const string TrafficLightSlow = "\uf639";
        public const string TrafficLightStop = "\uf63a";
        public const string Trailer = "\ue041";
        public const string Train = "\uf238";
        public const string Tram = "\uf7da";
        public const string Transgender = "\uf224";
        public const string TransgenderAlt = "\uf225";
        public const string Transporter = "\ue042";
        public const string Transporter1 = "\ue043";
        public const string Transporter2 = "\ue044";
        public const string Transporter3 = "\ue045";
        public const string TransporterEmpty = "\ue046";
        public const string Trash = "\uf1f8";
        public const string TrashAlt = "\uf2ed";
        public const string TrashRestore = "\uf829";
        public const string TrashRestoreAlt = "\uf82a";
        public const string TrashUndo = "\uf895";
        public const string TrashUndoAlt = "\uf896";
        public const string TreasureChest = "\uf723";
        public const string Tree = "\uf1bb";
        public const string TreeAlt = "\uf400";
        public const string TreeChristmas = "\uf7db";
        public const string TreeDecorated = "\uf7dc";
        public const string TreeLarge = "\uf7dd";
        public const string TreePalm = "\uf82b";
        public const string Trees = "\uf724";
        public const string Triangle = "\uf2ec";
        public const string TriangleMusic = "\uf8e2";
        public const string Trophy = "\uf091";
        public const string TrophyAlt = "\uf2eb";
        public const string Truck = "\uf0d1";
        public const string TruckContainer = "\uf4dc";
        public const string TruckCouch = "\uf4dd";
        public const string TruckLoading = "\uf4de";
        public const string TruckMonster = "\uf63b";
        public const string TruckMoving = "\uf4df";
        public const string TruckPickup = "\uf63c";
        public const string TruckPlow = "\uf7de";
        public const string TruckRamp = "\uf4e0";
        public const string Trumpet = "\uf8e3";
        public const string Tshirt = "\uf553";
        public const string Tty = "\uf1e4";
        public const string Turkey = "\uf725";
        public const string Turntable = "\uf8e4";
        public const string Turtle = "\uf726";
        public const string Tv = "\uf26c";
        public const string TvAlt = "\uf8e5";
        public const string TvMusic = "\uf8e6";
        public const string TvRetro = "\uf401";
        public const string Typewriter = "\uf8e7";
        public const string Ufo = "\ue047";
        public const string UfoBeam = "\ue048";
        public const string Umbrella = "\uf0e9";
        public const string UmbrellaBeach = "\uf5ca";
        public const string Underline = "\uf0cd";
        public const string Undo = "\uf0e2";
        public const string UndoAlt = "\uf2ea";
        public const string Unicorn = "\uf727";
        public const string Union = "\uf6a2";
        public const string UniversalAccess = "\uf29a";
        public const string University = "\uf19c";
        public const string Unlink = "\uf127";
        public const string Unlock = "\uf09c";
        public const string UnlockAlt = "\uf13e";
        public const string Upload = "\uf093";
        public const string UsbDrive = "\uf8e9";
        public const string UsdCircle = "\uf2e8";
        public const string UsdSquare = "\uf2e9";
        public const string User = "\uf007";
        public const string UserAlien = "\ue04a";
        public const string UserAlt = "\uf406";
        public const string UserAltSlash = "\uf4fa";
        public const string UserAstronaut = "\uf4fb";
        public const string UserChart = "\uf6a3";
        public const string UserCheck = "\uf4fc";
        public const string UserCircle = "\uf2bd";
        public const string UserClock = "\uf4fd";
        public const string UserCog = "\uf4fe";
        public const string UserCowboy = "\uf8ea";
        public const string UserCrown = "\uf6a4";
        public const string UserEdit = "\uf4ff";
        public const string UserFriends = "\uf500";
        public const string UserGraduate = "\uf501";
        public const string UserHardHat = "\uf82c";
        public const string UserHeadset = "\uf82d";
        public const string UserInjured = "\uf728";
        public const string UserLock = "\uf502";
        public const string UserMd = "\uf0f0";
        public const string UserMdChat = "\uf82e";
        public const string UserMinus = "\uf503";
        public const string UserMusic = "\uf8eb";
        public const string UserNinja = "\uf504";
        public const string UserNurse = "\uf82f";
        public const string UserPlus = "\uf234";
        public const string UserRobot = "\ue04b";
        public const string UserSecret = "\uf21b";
        public const string UserShield = "\uf505";
        public const string UserSlash = "\uf506";
        public const string UserTag = "\uf507";
        public const string UserTie = "\uf508";
        public const string UserTimes = "\uf235";
        public const string UserUnlock = "\ue058";
        public const string UserVisor = "\ue04c";
        public const string Users = "\uf0c0";
        public const string UsersClass = "\uf63d";
        public const string UsersCog = "\uf509";
        public const string UsersCrown = "\uf6a5";
        public const string UsersMedical = "\uf830";
        public const string UsersSlash = "\ue073";
        public const string UtensilFork = "\uf2e3";
        public const string UtensilKnife = "\uf2e4";
        public const string UtensilSpoon = "\uf2e5";
        public const string Utensils = "\uf2e7";
        public const string UtensilsAlt = "\uf2e6";
        public const string Vacuum = "\ue04d";
        public const string VacuumRobot = "\ue04e";
        public const string ValueAbsolute = "\uf6a6";
        public const string VectorSquare = "\uf5cb";
        public const string Venus = "\uf221";
        public const string VenusDouble = "\uf226";
        public const string VenusMars = "\uf228";
        public const string Vhs = "\uf8ec";
        public const string Vial = "\uf492";
        public const string Vials = "\uf493";
        public const string Video = "\uf03d";
        public const string VideoPlus = "\uf4e1";
        public const string VideoSlash = "\uf4e2";
        public const string Vihara = "\uf6a7";
        public const string Violin = "\uf8ed";
        public const string Virus = "\ue074";
        public const string VirusSlash = "\ue075";
        public const string Viruses = "\ue076";
        public const string Voicemail = "\uf897";
        public const string Volcano = "\uf770";
        public const string VolleyballBall = "\uf45f";
        public const string Volume = "\uf6a8";
        public const string VolumeDown = "\uf027";
        public const string VolumeMute = "\uf6a9";
        public const string VolumeOff = "\uf026";
        public const string VolumeSlash = "\uf2e2";
        public const string VolumeUp = "\uf028";
        public const string VoteNay = "\uf771";
        public const string VoteYea = "\uf772";
        public const string VrCardboard = "\uf729";
        public const string WagonCovered = "\uf8ee";
        public const string Walker = "\uf831";
        public const string WalkieTalkie = "\uf8ef";
        public const string Walking = "\uf554";
        public const string Wallet = "\uf555";
        public const string Wand = "\uf72a";
        public const string WandMagic = "\uf72b";
        public const string Warehouse = "\uf494";
        public const string WarehouseAlt = "\uf495";
        public const string Washer = "\uf898";
        public const string Watch = "\uf2e1";
        public const string WatchCalculator = "\uf8f0";
        public const string WatchFitness = "\uf63e";
        public const string Water = "\uf773";
        public const string WaterLower = "\uf774";
        public const string WaterRise = "\uf775";
        public const string WaveSine = "\uf899";
        public const string WaveSquare = "\uf83e";
        public const string WaveTriangle = "\uf89a";
        public const string Waveform = "\uf8f1";
        public const string WaveformPath = "\uf8f2";
        public const string Webcam = "\uf832";
        public const string WebcamSlash = "\uf833";
        public const string Weight = "\uf496";
        public const string WeightHanging = "\uf5cd";
        public const string Whale = "\uf72c";
        public const string Wheat = "\uf72d";
        public const string Wheelchair = "\uf193";
        public const string Whistle = "\uf460";
        public const string Wifi = "\uf1eb";
        public const string Wifi1 = "\uf6aa";
        public const string Wifi2 = "\uf6ab";
        public const string WifiSlash = "\uf6ac";
        public const string Wind = "\uf72e";
        public const string WindTurbine = "\uf89b";
        public const string WindWarning = "\uf776";
        public const string Window = "\uf40e";
        public const string WindowAlt = "\uf40f";
        public const string WindowClose = "\uf410";
        public const string WindowFrame = "\ue04f";
        public const string WindowFrameOpen = "\ue050";
        public const string WindowMaximize = "\uf2d0";
        public const string WindowMinimize = "\uf2d1";
        public const string WindowRestore = "\uf2d2";
        public const string Windsock = "\uf777";
        public const string WineBottle = "\uf72f";
        public const string WineGlass = "\uf4e3";
        public const string WineGlassAlt = "\uf5ce";
        public const string WonSign = "\uf159";
        public const string Wreath = "\uf7e2";
        public const string Wrench = "\uf0ad";
        public const string XRay = "\uf497";
        public const string YenSign = "\uf157";
        public const string YinYang = "\uf6ad";
    }
}
