// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C#
// from https://raw.githubusercontent.com/fefanto/fontaudio/master/font/fontaudio.css
// for use with https://github.com/fefanto/fontaudio/raw/master/font/fontaudio.ttf
namespace IconFonts
{
    public class Fontaudio
    {
        public const string FontIconFileName = "fontaudio.ttf";

        public const int IconMin = 0xf101;
        public const int IconMax = 0xf19b;
        public const string Adr = "\uf101";
        public const string Adsr = "\uf102";
        public const string Ahdsr = "\uf103";
        public const string Ar = "\uf104";
        public const string Armrecording = "\uf105";
        public const string Arpchord = "\uf106";
        public const string Arpdown = "\uf107";
        public const string Arpdownandup = "\uf108";
        public const string Arpdownup = "\uf109";
        public const string Arpplayorder = "\uf10a";
        public const string Arprandom = "\uf10b";
        public const string Arpup = "\uf10c";
        public const string Arpupandown = "\uf10d";
        public const string Arpupdown = "\uf10e";
        public const string ArrowsHorz = "\uf10f";
        public const string ArrowsVert = "\uf110";
        public const string Automation2p = "\uf111";
        public const string Automation3p = "\uf112";
        public const string Automation4p = "\uf113";
        public const string Backward = "\uf114";
        public const string Bluetooth = "\uf115";
        public const string CaretDown = "\uf116";
        public const string CaretLeft = "\uf117";
        public const string CaretRight = "\uf118";
        public const string CaretUp = "\uf119";
        public const string Close = "\uf11a";
        public const string Copy = "\uf11b";
        public const string Cpu = "\uf11c";
        public const string Cutter = "\uf11d";
        public const string DigitalColon = "\uf11e";
        public const string DigitalDot = "\uf11f";
        public const string Digital0 = "\uf120";
        public const string Digital1 = "\uf121";
        public const string Digital2 = "\uf122";
        public const string Digital3 = "\uf123";
        public const string Digital4 = "\uf124";
        public const string Digital5 = "\uf125";
        public const string Digital6 = "\uf126";
        public const string Digital7 = "\uf127";
        public const string Digital8 = "\uf128";
        public const string Digital9 = "\uf129";
        public const string Diskio = "\uf12a";
        public const string Drumpad = "\uf12b";
        public const string Duplicate = "\uf12c";
        public const string Eraser = "\uf12d";
        public const string Ffwd = "\uf12e";
        public const string FilterBandpass = "\uf12f";
        public const string FilterBell = "\uf130";
        public const string FilterBypass = "\uf131";
        public const string FilterHighpass = "\uf132";
        public const string FilterLowpass = "\uf133";
        public const string FilterNotch = "\uf134";
        public const string FilterRezHighpass = "\uf135";
        public const string FilterRezLowpass = "\uf136";
        public const string FilterShelvingHi = "\uf137";
        public const string FilterShelvingLo = "\uf138";
        public const string Foldback = "\uf139";
        public const string Forward = "\uf13a";
        public const string HExpand = "\uf13b";
        public const string Hardclip = "\uf13c";
        public const string Hardclipcurve = "\uf13d";
        public const string Headphones = "\uf13e";
        public const string Keyboard = "\uf13f";
        public const string Lock = "\uf140";
        public const string LogoAax = "\uf141";
        public const string LogoAbletonlink = "\uf142";
        public const string LogoAu = "\uf143";
        public const string LogoAudacity = "\uf144";
        public const string LogoAudiobus = "\uf145";
        public const string LogoCubase = "\uf146";
        public const string LogoFl = "\uf147";
        public const string LogoJuce = "\uf148";
        public const string LogoLadspa = "\uf149";
        public const string LogoLive = "\uf14a";
        public const string LogoLv2 = "\uf14b";
        public const string LogoProtools = "\uf14c";
        public const string LogoRackext = "\uf14d";
        public const string LogoReaper = "\uf14e";
        public const string LogoReason = "\uf14f";
        public const string LogoRewire = "\uf150";
        public const string LogoStudioone = "\uf151";
        public const string LogoTracktion = "\uf152";
        public const string LogoVst = "\uf153";
        public const string LogoWaveform = "\uf154";
        public const string Loop = "\uf155";
        public const string Metronome = "\uf156";
        public const string Microphone = "\uf157";
        public const string Midiplug = "\uf158";
        public const string Modrandom = "\uf159";
        public const string Modsawdown = "\uf15a";
        public const string Modsawup = "\uf15b";
        public const string Modsh = "\uf15c";
        public const string Modsine = "\uf15d";
        public const string Modsquare = "\uf15e";
        public const string Modtri = "\uf15f";
        public const string Modularplug = "\uf160";
        public const string Mono = "\uf161";
        public const string Mute = "\uf162";
        public const string Next = "\uf163";
        public const string Open = "\uf164";
        public const string Paste = "\uf165";
        public const string Pause = "\uf166";
        public const string Pen = "\uf167";
        public const string Phase = "\uf168";
        public const string Play = "\uf169";
        public const string Pointer = "\uf16a";
        public const string Powerswitch = "\uf16b";
        public const string PresetA = "\uf16c";
        public const string PresetAb = "\uf16d";
        public const string PresetB = "\uf16e";
        public const string PresetBa = "\uf16f";
        public const string Prev = "\uf170";
        public const string PunchIn = "\uf171";
        public const string PunchOut = "\uf172";
        public const string Ram = "\uf173";
        public const string Random1dice = "\uf174";
        public const string Random2dice = "\uf175";
        public const string Record = "\uf176";
        public const string Redo = "\uf177";
        public const string RepeatOne = "\uf178";
        public const string Repeat = "\uf179";
        public const string Rew = "\uf17a";
        public const string RoundswitchOff = "\uf17b";
        public const string RoundswitchOn = "\uf17c";
        public const string Save = "\uf17d";
        public const string Saveas = "\uf17e";
        public const string Scissors = "\uf17f";
        public const string Shuffle = "\uf180";
        public const string SliderRound1 = "\uf181";
        public const string SliderRound2 = "\uf182";
        public const string SliderRound3 = "\uf183";
        public const string Sliderhandle1 = "\uf184";
        public const string Sliderhandle2 = "\uf185";
        public const string Softclip = "\uf186";
        public const string Softclipcurve = "\uf187";
        public const string Solo = "\uf188";
        public const string Speaker = "\uf189";
        public const string SquareswitchOff = "\uf18a";
        public const string SquareswitchOn = "\uf18b";
        public const string Stereo = "\uf18c";
        public const string Stop = "\uf18d";
        public const string Thunderbolt = "\uf18e";
        public const string Timeselect = "\uf18f";
        public const string Undo = "\uf190";
        public const string Unlock = "\uf191";
        public const string Usb = "\uf192";
        public const string VExpand = "\uf193";
        public const string VroundswitchOff = "\uf194";
        public const string VroundswitchOn = "\uf195";
        public const string VsquareswitchOff = "\uf196";
        public const string VsquareswitchOn = "\uf197";
        public const string Waveform = "\uf198";
        public const string Xlrplug = "\uf199";
        public const string Zoomin = "\uf19a";
        public const string Zoomout = "\uf19b";
    }
}
