// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C++11
// from https://raw.githubusercontent.com/nicodinh/kenney-icon-font/master/css/kenney-icons.css
// for use with https://github.com/nicodinh/kenney-icon-font/blob/master/fonts/kenney-icon-font.ttf
#pragma once

#define FONT_ICON_FILE_NAME_KI "kenney-icon-font.ttf"

#define ICON_MIN_KI 0xe900
#define ICON_MAX_KI 0xe9e3
#define ICON_KI_HOME u8"\ue900"
#define ICON_KI_ADJUST u8"\ue901"
#define ICON_KI_WRENCH u8"\ue902"
#define ICON_KI_COG u8"\ue903"
#define ICON_KI_OFF u8"\ue904"
#define ICON_KI_EXPAND u8"\ue905"
#define ICON_KI_REDUCE u8"\ue906"
#define ICON_KI_MOVIE u8"\ue907"
#define ICON_KI_FLAP u8"\ue908"
#define ICON_KI_SHOPPING_CART u8"\ue909"
#define ICON_KI_SHOPPING_CASE u8"\ue90a"
#define ICON_KI_EXTERNAL u8"\ue90b"
#define ICON_KI_NETWORK u8"\ue90c"
#define ICON_KI_CHECK u8"\ue90d"
#define ICON_KI_TIMES u8"\ue90e"
#define ICON_KI_TIMES_CIRCLE u8"\ue90f"
#define ICON_KI_PLUS u8"\ue910"
#define ICON_KI_PLUS_CIRCLE u8"\ue911"
#define ICON_KI_MINUS u8"\ue912"
#define ICON_KI_MINUS_CIRCLE u8"\ue913"
#define ICON_KI_INFO u8"\ue914"
#define ICON_KI_INFO_CIRCLE u8"\ue915"
#define ICON_KI_QUESTION u8"\ue916"
#define ICON_KI_QUESTION_CIRCLE u8"\ue917"
#define ICON_KI_EXLAMATION u8"\ue918"
#define ICON_KI_EXCLAMATION_CIRCLE u8"\ue919"
#define ICON_KI_EXCLAMATION_TRIANGLE u8"\ue91a"
#define ICON_KI_PAINT_BRUSH u8"\ue91b"
#define ICON_KI_PENCIL u8"\ue91c"
#define ICON_KI_CHECKBOX u8"\ue91d"
#define ICON_KI_CHECKBOX_CHECKED u8"\ue91e"
#define ICON_KI_RADIO u8"\ue91f"
#define ICON_KI_RADIO_CHECKED u8"\ue920"
#define ICON_KI_SORT_VERTICAL u8"\ue921"
#define ICON_KI_SORT_HORIZONTAL u8"\ue922"
#define ICON_KI_GRID u8"\ue923"
#define ICON_KI_LIST u8"\ue924"
#define ICON_KI_ROWS u8"\ue925"
#define ICON_KI_CELLS u8"\ue926"
#define ICON_KI_SIGNAL_LOW u8"\ue927"
#define ICON_KI_SIGNAL_MEDIUM u8"\ue928"
#define ICON_KI_SIGNAL_HIGH u8"\ue929"
#define ICON_KI_TRASH u8"\ue92a"
#define ICON_KI_TRASH_ALT u8"\ue92b"
#define ICON_KI_RELOAD_INVERSE u8"\ue92c"
#define ICON_KI_RELOAD u8"\ue92d"
#define ICON_KI_TOP u8"\ue92e"
#define ICON_KI_BOTTOM u8"\ue92f"
#define ICON_KI_UPLOAD u8"\ue930"
#define ICON_KI_DOWNLOAD u8"\ue931"
#define ICON_KI_CLOUD u8"\ue932"
#define ICON_KI_CLOUD_UPLOAD u8"\ue933"
#define ICON_KI_CLOUD_DOWNLOAD u8"\ue934"
#define ICON_KI_SEARCH u8"\ue935"
#define ICON_KI_SEARCH_PLUS u8"\ue936"
#define ICON_KI_SEARCH_MINUS u8"\ue937"
#define ICON_KI_SEARCH_EQUAL u8"\ue938"
#define ICON_KI_LOCK u8"\ue939"
#define ICON_KI_UNLOCK u8"\ue93a"
#define ICON_KI_USER u8"\ue93b"
#define ICON_KI_USERS u8"\ue93c"
#define ICON_KI_USERS_ALT u8"\ue93d"
#define ICON_KI_SIGN_IN u8"\ue93e"
#define ICON_KI_SIGN_IN_INVERSE u8"\ue93f"
#define ICON_KI_SIGN_OUT u8"\ue940"
#define ICON_KI_SIGN_OUT_INVERSE u8"\ue941"
#define ICON_KI_ARROW_TOP u8"\ue942"
#define ICON_KI_ARROW_RIGHT u8"\ue943"
#define ICON_KI_ARROW_BOTTOM u8"\ue944"
#define ICON_KI_ARROW_LEFT u8"\ue945"
#define ICON_KI_ARROW_TOP_LEFT u8"\ue946"
#define ICON_KI_ARROW_TOP_RIGHT u8"\ue947"
#define ICON_KI_ARROW_BOTTOM_RIGHT u8"\ue948"
#define ICON_KI_ARROW_BOTTOM_LEFT u8"\ue949"
#define ICON_KI_CARET_TOP u8"\ue94a"
#define ICON_KI_CARET_RIGHT u8"\ue94b"
#define ICON_KI_CARET_BOTTOM u8"\ue94c"
#define ICON_KI_CARET_LEFT u8"\ue94d"
#define ICON_KI_NEXT_ALT u8"\ue94e"
#define ICON_KI_NEXT u8"\ue94f"
#define ICON_KI_PREVIOUS u8"\ue950"
#define ICON_KI_PREVIOUS_ALT u8"\ue951"
#define ICON_KI_FILL u8"\ue952"
#define ICON_KI_ERASER u8"\ue953"
#define ICON_KI_SAVE u8"\ue954"
#define ICON_KI_STEP_BACKWARD u8"\ue955"
#define ICON_KI_BACKWARD u8"\ue956"
#define ICON_KI_PAUSE u8"\ue957"
#define ICON_KI_FORWARD u8"\ue958"
#define ICON_KI_STEP_FORWARD u8"\ue959"
#define ICON_KI_STOP u8"\ue95a"
#define ICON_KI_REC u8"\ue95b"
#define ICON_KI_CURSOR u8"\ue95c"
#define ICON_KI_POINTER u8"\ue95d"
#define ICON_KI_EXIT u8"\ue95e"
#define ICON_KI_FIGURE u8"\ue95f"
#define ICON_KI_CAR u8"\ue960"
#define ICON_KI_COIN u8"\ue961"
#define ICON_KI_KEY u8"\ue962"
#define ICON_KI_CUB u8"\ue963"
#define ICON_KI_DIAMOND u8"\ue964"
#define ICON_KI_BADGE u8"\ue965"
#define ICON_KI_BADGE_ALT u8"\ue966"
#define ICON_KI_PODIUM u8"\ue967"
#define ICON_KI_PODIUM_ALT u8"\ue968"
#define ICON_KI_FLAG u8"\ue969"
#define ICON_KI_FIST u8"\ue96a"
#define ICON_KI_FIST_CIRCLE u8"\ue96b"
#define ICON_KI_HEART u8"\ue96c"
#define ICON_KI_HEART_HALF u8"\ue96d"
#define ICON_KI_HEART_HALF_O u8"\ue96e"
#define ICON_KI_HEART_O u8"\ue96f"
#define ICON_KI_STAR u8"\ue970"
#define ICON_KI_STAR_HALF u8"\ue971"
#define ICON_KI_STAR_HALF_O u8"\ue972"
#define ICON_KI_STAR_O u8"\ue973"
#define ICON_KI_BUTTON_B u8"\ue974"
#define ICON_KI_MUSIC_ON u8"\ue975"
#define ICON_KI_MUSIC_OFF u8"\ue976"
#define ICON_KI_SOUND_ON u8"\ue977"
#define ICON_KI_SOUND_OFF u8"\ue978"
#define ICON_KI_SOUND_OFF_ALT u8"\ue979"
#define ICON_KI_ROBOT u8"\ue97a"
#define ICON_KI_COMPUTER u8"\ue97b"
#define ICON_KI_TABLET u8"\ue97c"
#define ICON_KI_SMARTPHONE u8"\ue97d"
#define ICON_KI_DEVICE u8"\ue97e"
#define ICON_KI_DEVICE_TILT_LEFT u8"\ue97f"
#define ICON_KI_DEVICE_TILT_RIGHT u8"\ue980"
#define ICON_KI_GAMEPAD u8"\ue981"
#define ICON_KI_GAMEPAD_ALT u8"\ue982"
#define ICON_KI_GAMEPAD_TILT_LEFT u8"\ue983"
#define ICON_KI_GAMEPAD_TILT_RIGHT u8"\ue984"
#define ICON_KI_PLAYER_ONE u8"\ue985"
#define ICON_KI_PLAYER_TWO u8"\ue986"
#define ICON_KI_PLAYER_THREE u8"\ue987"
#define ICON_KI_PLAYER_FOUR u8"\ue988"
#define ICON_KI_JOYSTICK u8"\ue989"
#define ICON_KI_JOYSTICK_ALT u8"\ue98a"
#define ICON_KI_JOYSTICK_LEFT u8"\ue98b"
#define ICON_KI_JOYSTICK_RIGHT u8"\ue98c"
#define ICON_KI_MOUSE_ALT u8"\ue98d"
#define ICON_KI_MOUSE u8"\ue98e"
#define ICON_KI_MOUSE_LEFT_BUTTON u8"\ue98f"
#define ICON_KI_MOUSE_RIGHT_BUTTON u8"\ue990"
#define ICON_KI_BUTTON_ONE u8"\ue991"
#define ICON_KI_BUTTON_TWO u8"\ue992"
#define ICON_KI_BUTTON_THREE u8"\ue993"
#define ICON_KI_BUTTON_A u8"\ue994"
#define ICON_KI_BUTTON_X u8"\ue995"
#define ICON_KI_BUTON_Y u8"\ue996"
#define ICON_KI_BUTTON_TIMES u8"\ue997"
#define ICON_KI_BUTTON_SQUARE u8"\ue998"
#define ICON_KI_BUTTON_CIRCLE u8"\ue999"
#define ICON_KI_BUTTON_TRIANGLE u8"\ue99a"
#define ICON_KI_BUTTON_LEFT u8"\ue99b"
#define ICON_KI_BUTTON_L u8"\ue99c"
#define ICON_KI_BUTTON_L1 u8"\ue99d"
#define ICON_KI_BUTTON_L2 u8"\ue99e"
#define ICON_KI_BUTTON_LB u8"\ue99f"
#define ICON_KI_BUTTON_LT u8"\ue9a0"
#define ICON_KI_BUTTON_RT u8"\ue9a1"
#define ICON_KI_BUTTON_RB u8"\ue9a2"
#define ICON_KI_BUTTON_R2 u8"\ue9a3"
#define ICON_KI_BUTTON_R1 u8"\ue9a4"
#define ICON_KI_BUTTON_R u8"\ue9a5"
#define ICON_KI_BUTTON_RIGHT u8"\ue9a6"
#define ICON_KI_BUTTON_EMPTY u8"\ue9a7"
#define ICON_KI_BUTTON_START u8"\ue9a8"
#define ICON_KI_BUTTON_SELECT u8"\ue9a9"
#define ICON_KI_DPAD u8"\ue9aa"
#define ICON_KI_DPAD_ALT u8"\ue9ab"
#define ICON_KI_DPAD_TOP u8"\ue9ac"
#define ICON_KI_DPAD_RIGHT u8"\ue9ad"
#define ICON_KI_DPAD_BOTTOM u8"\ue9ae"
#define ICON_KI_DPAD_LEFT u8"\ue9af"
#define ICON_KI_KEY_LARGE u8"\ue9b0"
#define ICON_KI_KEY_LARGE_3D u8"\ue9b1"
#define ICON_KI_KEY_SMALL u8"\ue9b2"
#define ICON_KI_KEY_SMALL_3D u8"\ue9b3"
#define ICON_KI_STICK_LEFT_TOP u8"\ue9b4"
#define ICON_KI_STICK_LEFT_SIDE u8"\ue9b5"
#define ICON_KI_STICK_RIGHT_SIDE u8"\ue9b6"
#define ICON_KI_STICK_RIGHT_TOP u8"\ue9b7"
#define ICON_KI_STICK_SIDE u8"\ue9b8"
#define ICON_KI_STICK_TILT_LEFT u8"\ue9b9"
#define ICON_KI_STICK_TILT_RIGHT u8"\ue9ba"
#define ICON_KI_MOVE_BL u8"\ue9bb"
#define ICON_KI_MOVE_BR u8"\ue9bc"
#define ICON_KI_MOVE_BT u8"\ue9bd"
#define ICON_KI_MOVE_BT_ALT u8"\ue9be"
#define ICON_KI_MOVE_LB u8"\ue9bf"
#define ICON_KI_MOVE_LR u8"\ue9c0"
#define ICON_KI_MOVE_LR_ALT u8"\ue9c1"
#define ICON_KI_MOVE_LT u8"\ue9c2"
#define ICON_KI_MOVE_RB u8"\ue9c3"
#define ICON_KI_MOVE_RL u8"\ue9c4"
#define ICON_KI_MOVE_RL_ALT u8"\ue9c5"
#define ICON_KI_MOVE_RT u8"\ue9c6"
#define ICON_KI_MOVE_TB u8"\ue9c7"
#define ICON_KI_MOVE_TB_ALT u8"\ue9c8"
#define ICON_KI_MOVE_TL u8"\ue9c9"
#define ICON_KI_MOVE_TR u8"\ue9ca"
#define ICON_KI_STICK_MOVE_BL u8"\ue9cb"
#define ICON_KI_STICK_MOVE_BR u8"\ue9cc"
#define ICON_KI_STICK_MOVE_BT u8"\ue9cd"
#define ICON_KI_STICK_MOVE_BT_ALT u8"\ue9ce"
#define ICON_KI_STICK_MOVE_LB u8"\ue9cf"
#define ICON_KI_STICK_MOVE_LR u8"\ue9d0"
#define ICON_KI_STICK_MOVE_LR_ALT u8"\ue9d1"
#define ICON_KI_STICK_MOVE_LT u8"\ue9d2"
#define ICON_KI_STICK_MOVE_RB u8"\ue9d3"
#define ICON_KI_STICK_MOVE_RL u8"\ue9d4"
#define ICON_KI_STICK_MOVE_RL_ALT u8"\ue9d5"
#define ICON_KI_STICK_MOVE_RT u8"\ue9d6"
#define ICON_KI_STICK_MOVE_TB u8"\ue9d7"
#define ICON_KI_STICK_MOVE_TB_ALT u8"\ue9d8"
#define ICON_KI_STICK_MOVE_TL u8"\ue9d9"
#define ICON_KI_STICK_MOVE_TR u8"\ue9da"
#define ICON_KI_GITHUB u8"\ue9db"
#define ICON_KI_GITHUB_ALT u8"\ue9dc"
#define ICON_KI_TWITTER u8"\ue9dd"
#define ICON_KI_FACEBOOK u8"\ue9de"
#define ICON_KI_GOOGLE_PLUS u8"\ue9df"
#define ICON_KI_YOUTUBE u8"\ue9e2"
#define ICON_KI_WE_HEART u8"\ue9e3"
#define ICON_KI_WOLFCMS u8"\ue9e0"
#define ICON_KI_WOLFCMS_ALT u8"\ue9e1"
