// Generated by https://github.com/juliettef/IconFontCppHeaders script GenerateIconFontCppHeaders.py for language C#
// from https://raw.githubusercontent.com/google/material-design-icons/master/iconfont/codepoints
// for use with https://github.com/google/material-design-icons/blob/master/iconfont/MaterialIcons-Regular.ttf
namespace IconFonts
{
    public class MaterialDesign
    {
        public const string FontIconFileName = "MaterialIcons-Regular.ttf";

        public const int IconMin = 0xe000;
        public const int IconMax = 0xeb4c;
        public const string The3d_rotation = "\ue84d";
        public const string Ac_unit = "\ueb3b";
        public const string Access_alarm = "\ue190";
        public const string Access_alarms = "\ue191";
        public const string Access_time = "\ue192";
        public const string Accessibility = "\ue84e";
        public const string Accessible = "\ue914";
        public const string Account_balance = "\ue84f";
        public const string Account_balance_wallet = "\ue850";
        public const string Account_box = "\ue851";
        public const string Account_circle = "\ue853";
        public const string Adb = "\ue60e";
        public const string Add = "\ue145";
        public const string Add_a_photo = "\ue439";
        public const string Add_alarm = "\ue193";
        public const string Add_alert = "\ue003";
        public const string Add_box = "\ue146";
        public const string Add_circle = "\ue147";
        public const string Add_circle_outline = "\ue148";
        public const string Add_location = "\ue567";
        public const string Add_shopping_cart = "\ue854";
        public const string Add_to_photos = "\ue39d";
        public const string Add_to_queue = "\ue05c";
        public const string Adjust = "\ue39e";
        public const string Airline_seat_flat = "\ue630";
        public const string Airline_seat_flat_angled = "\ue631";
        public const string Airline_seat_individual_suite = "\ue632";
        public const string Airline_seat_legroom_extra = "\ue633";
        public const string Airline_seat_legroom_normal = "\ue634";
        public const string Airline_seat_legroom_reduced = "\ue635";
        public const string Airline_seat_recline_extra = "\ue636";
        public const string Airline_seat_recline_normal = "\ue637";
        public const string Airplanemode_active = "\ue195";
        public const string Airplanemode_inactive = "\ue194";
        public const string Airplay = "\ue055";
        public const string Airport_shuttle = "\ueb3c";
        public const string Alarm = "\ue855";
        public const string Alarm_add = "\ue856";
        public const string Alarm_off = "\ue857";
        public const string Alarm_on = "\ue858";
        public const string Album = "\ue019";
        public const string All_inclusive = "\ueb3d";
        public const string All_out = "\ue90b";
        public const string Android = "\ue859";
        public const string Announcement = "\ue85a";
        public const string Apps = "\ue5c3";
        public const string Archive = "\ue149";
        public const string Arrow_back = "\ue5c4";
        public const string Arrow_downward = "\ue5db";
        public const string Arrow_drop_down = "\ue5c5";
        public const string Arrow_drop_down_circle = "\ue5c6";
        public const string Arrow_drop_up = "\ue5c7";
        public const string Arrow_forward = "\ue5c8";
        public const string Arrow_upward = "\ue5d8";
        public const string Art_track = "\ue060";
        public const string Aspect_ratio = "\ue85b";
        public const string Assessment = "\ue85c";
        public const string Assignment = "\ue85d";
        public const string Assignment_ind = "\ue85e";
        public const string Assignment_late = "\ue85f";
        public const string Assignment_return = "\ue860";
        public const string Assignment_returned = "\ue861";
        public const string Assignment_turned_in = "\ue862";
        public const string Assistant = "\ue39f";
        public const string Assistant_photo = "\ue3a0";
        public const string Attach_file = "\ue226";
        public const string Attach_money = "\ue227";
        public const string Attachment = "\ue2bc";
        public const string Audiotrack = "\ue3a1";
        public const string Autorenew = "\ue863";
        public const string Av_timer = "\ue01b";
        public const string Backspace = "\ue14a";
        public const string Backup = "\ue864";
        public const string Battery_alert = "\ue19c";
        public const string Battery_charging_full = "\ue1a3";
        public const string Battery_full = "\ue1a4";
        public const string Battery_std = "\ue1a5";
        public const string Battery_unknown = "\ue1a6";
        public const string Beach_access = "\ueb3e";
        public const string Beenhere = "\ue52d";
        public const string Block = "\ue14b";
        public const string Bluetooth = "\ue1a7";
        public const string Bluetooth_audio = "\ue60f";
        public const string Bluetooth_connected = "\ue1a8";
        public const string Bluetooth_disabled = "\ue1a9";
        public const string Bluetooth_searching = "\ue1aa";
        public const string Blur_circular = "\ue3a2";
        public const string Blur_linear = "\ue3a3";
        public const string Blur_off = "\ue3a4";
        public const string Blur_on = "\ue3a5";
        public const string Book = "\ue865";
        public const string Bookmark = "\ue866";
        public const string Bookmark_border = "\ue867";
        public const string Border_all = "\ue228";
        public const string Border_bottom = "\ue229";
        public const string Border_clear = "\ue22a";
        public const string Border_color = "\ue22b";
        public const string Border_horizontal = "\ue22c";
        public const string Border_inner = "\ue22d";
        public const string Border_left = "\ue22e";
        public const string Border_outer = "\ue22f";
        public const string Border_right = "\ue230";
        public const string Border_style = "\ue231";
        public const string Border_top = "\ue232";
        public const string Border_vertical = "\ue233";
        public const string Branding_watermark = "\ue06b";
        public const string Brightness_1 = "\ue3a6";
        public const string Brightness_2 = "\ue3a7";
        public const string Brightness_3 = "\ue3a8";
        public const string Brightness_4 = "\ue3a9";
        public const string Brightness_5 = "\ue3aa";
        public const string Brightness_6 = "\ue3ab";
        public const string Brightness_7 = "\ue3ac";
        public const string Brightness_auto = "\ue1ab";
        public const string Brightness_high = "\ue1ac";
        public const string Brightness_low = "\ue1ad";
        public const string Brightness_medium = "\ue1ae";
        public const string Broken_image = "\ue3ad";
        public const string Brush = "\ue3ae";
        public const string Bubble_chart = "\ue6dd";
        public const string Bug_report = "\ue868";
        public const string Build = "\ue869";
        public const string Burst_mode = "\ue43c";
        public const string Business = "\ue0af";
        public const string Business_center = "\ueb3f";
        public const string Cached = "\ue86a";
        public const string Cake = "\ue7e9";
        public const string Call = "\ue0b0";
        public const string Call_end = "\ue0b1";
        public const string Call_made = "\ue0b2";
        public const string Call_merge = "\ue0b3";
        public const string Call_missed = "\ue0b4";
        public const string Call_missed_outgoing = "\ue0e4";
        public const string Call_received = "\ue0b5";
        public const string Call_split = "\ue0b6";
        public const string Call_to_action = "\ue06c";
        public const string Camera = "\ue3af";
        public const string Camera_alt = "\ue3b0";
        public const string Camera_enhance = "\ue8fc";
        public const string Camera_front = "\ue3b1";
        public const string Camera_rear = "\ue3b2";
        public const string Camera_roll = "\ue3b3";
        public const string Cancel = "\ue5c9";
        public const string Card_giftcard = "\ue8f6";
        public const string Card_membership = "\ue8f7";
        public const string Card_travel = "\ue8f8";
        public const string Casino = "\ueb40";
        public const string Cast = "\ue307";
        public const string Cast_connected = "\ue308";
        public const string Center_focus_strong = "\ue3b4";
        public const string Center_focus_weak = "\ue3b5";
        public const string Change_history = "\ue86b";
        public const string Chat = "\ue0b7";
        public const string Chat_bubble = "\ue0ca";
        public const string Chat_bubble_outline = "\ue0cb";
        public const string Check = "\ue5ca";
        public const string Check_box = "\ue834";
        public const string Check_box_outline_blank = "\ue835";
        public const string Check_circle = "\ue86c";
        public const string Chevron_left = "\ue5cb";
        public const string Chevron_right = "\ue5cc";
        public const string Child_care = "\ueb41";
        public const string Child_friendly = "\ueb42";
        public const string Chrome_reader_mode = "\ue86d";
        public const string Class = "\ue86e";
        public const string Clear = "\ue14c";
        public const string Clear_all = "\ue0b8";
        public const string Close = "\ue5cd";
        public const string Closed_caption = "\ue01c";
        public const string Cloud = "\ue2bd";
        public const string Cloud_circle = "\ue2be";
        public const string Cloud_done = "\ue2bf";
        public const string Cloud_download = "\ue2c0";
        public const string Cloud_off = "\ue2c1";
        public const string Cloud_queue = "\ue2c2";
        public const string Cloud_upload = "\ue2c3";
        public const string Code = "\ue86f";
        public const string Collections = "\ue3b6";
        public const string Collections_bookmark = "\ue431";
        public const string Color_lens = "\ue3b7";
        public const string Colorize = "\ue3b8";
        public const string Comment = "\ue0b9";
        public const string Compare = "\ue3b9";
        public const string Compare_arrows = "\ue915";
        public const string Computer = "\ue30a";
        public const string Confirmation_number = "\ue638";
        public const string Contact_mail = "\ue0d0";
        public const string Contact_phone = "\ue0cf";
        public const string Contacts = "\ue0ba";
        public const string Content_copy = "\ue14d";
        public const string Content_cut = "\ue14e";
        public const string Content_paste = "\ue14f";
        public const string Control_point = "\ue3ba";
        public const string Control_point_duplicate = "\ue3bb";
        public const string Copyright = "\ue90c";
        public const string Create = "\ue150";
        public const string Create_new_folder = "\ue2cc";
        public const string Credit_card = "\ue870";
        public const string Crop = "\ue3be";
        public const string Crop_16_9 = "\ue3bc";
        public const string Crop_3_2 = "\ue3bd";
        public const string Crop_5_4 = "\ue3bf";
        public const string Crop_7_5 = "\ue3c0";
        public const string Crop_din = "\ue3c1";
        public const string Crop_free = "\ue3c2";
        public const string Crop_landscape = "\ue3c3";
        public const string Crop_original = "\ue3c4";
        public const string Crop_portrait = "\ue3c5";
        public const string Crop_rotate = "\ue437";
        public const string Crop_square = "\ue3c6";
        public const string Dashboard = "\ue871";
        public const string Data_usage = "\ue1af";
        public const string Date_range = "\ue916";
        public const string Dehaze = "\ue3c7";
        public const string Delete = "\ue872";
        public const string Delete_forever = "\ue92b";
        public const string Delete_sweep = "\ue16c";
        public const string Description = "\ue873";
        public const string Desktop_mac = "\ue30b";
        public const string Desktop_windows = "\ue30c";
        public const string Details = "\ue3c8";
        public const string Developer_board = "\ue30d";
        public const string Developer_mode = "\ue1b0";
        public const string Device_hub = "\ue335";
        public const string Devices = "\ue1b1";
        public const string Devices_other = "\ue337";
        public const string Dialer_sip = "\ue0bb";
        public const string Dialpad = "\ue0bc";
        public const string Directions = "\ue52e";
        public const string Directions_bike = "\ue52f";
        public const string Directions_boat = "\ue532";
        public const string Directions_bus = "\ue530";
        public const string Directions_car = "\ue531";
        public const string Directions_railway = "\ue534";
        public const string Directions_run = "\ue566";
        public const string Directions_subway = "\ue533";
        public const string Directions_transit = "\ue535";
        public const string Directions_walk = "\ue536";
        public const string Disc_full = "\ue610";
        public const string Dns = "\ue875";
        public const string Do_not_disturb = "\ue612";
        public const string Do_not_disturb_alt = "\ue611";
        public const string Do_not_disturb_off = "\ue643";
        public const string Do_not_disturb_on = "\ue644";
        public const string Dock = "\ue30e";
        public const string Domain = "\ue7ee";
        public const string Done = "\ue876";
        public const string Done_all = "\ue877";
        public const string Donut_large = "\ue917";
        public const string Donut_small = "\ue918";
        public const string Drafts = "\ue151";
        public const string Drag_handle = "\ue25d";
        public const string Drive_eta = "\ue613";
        public const string Dvr = "\ue1b2";
        public const string Edit = "\ue3c9";
        public const string Edit_location = "\ue568";
        public const string Eject = "\ue8fb";
        public const string Email = "\ue0be";
        public const string Enhanced_encryption = "\ue63f";
        public const string Equalizer = "\ue01d";
        public const string Error = "\ue000";
        public const string Error_outline = "\ue001";
        public const string Euro_symbol = "\ue926";
        public const string Ev_station = "\ue56d";
        public const string Event = "\ue878";
        public const string Event_available = "\ue614";
        public const string Event_busy = "\ue615";
        public const string Event_note = "\ue616";
        public const string Event_seat = "\ue903";
        public const string Exit_to_app = "\ue879";
        public const string Expand_less = "\ue5ce";
        public const string Expand_more = "\ue5cf";
        public const string Explicit = "\ue01e";
        public const string Explore = "\ue87a";
        public const string Exposure = "\ue3ca";
        public const string Exposure_neg_1 = "\ue3cb";
        public const string Exposure_neg_2 = "\ue3cc";
        public const string Exposure_plus_1 = "\ue3cd";
        public const string Exposure_plus_2 = "\ue3ce";
        public const string Exposure_zero = "\ue3cf";
        public const string Extension = "\ue87b";
        public const string Face = "\ue87c";
        public const string Fast_forward = "\ue01f";
        public const string Fast_rewind = "\ue020";
        public const string Favorite = "\ue87d";
        public const string Favorite_border = "\ue87e";
        public const string Featured_play_list = "\ue06d";
        public const string Featured_video = "\ue06e";
        public const string Feedback = "\ue87f";
        public const string Fiber_dvr = "\ue05d";
        public const string Fiber_manual_record = "\ue061";
        public const string Fiber_new = "\ue05e";
        public const string Fiber_pin = "\ue06a";
        public const string Fiber_smart_record = "\ue062";
        public const string File_download = "\ue2c4";
        public const string File_upload = "\ue2c6";
        public const string Filter = "\ue3d3";
        public const string Filter_1 = "\ue3d0";
        public const string Filter_2 = "\ue3d1";
        public const string Filter_3 = "\ue3d2";
        public const string Filter_4 = "\ue3d4";
        public const string Filter_5 = "\ue3d5";
        public const string Filter_6 = "\ue3d6";
        public const string Filter_7 = "\ue3d7";
        public const string Filter_8 = "\ue3d8";
        public const string Filter_9 = "\ue3d9";
        public const string Filter_9_plus = "\ue3da";
        public const string Filter_b_and_w = "\ue3db";
        public const string Filter_center_focus = "\ue3dc";
        public const string Filter_drama = "\ue3dd";
        public const string Filter_frames = "\ue3de";
        public const string Filter_hdr = "\ue3df";
        public const string Filter_list = "\ue152";
        public const string Filter_none = "\ue3e0";
        public const string Filter_tilt_shift = "\ue3e2";
        public const string Filter_vintage = "\ue3e3";
        public const string Find_in_page = "\ue880";
        public const string Find_replace = "\ue881";
        public const string Fingerprint = "\ue90d";
        public const string First_page = "\ue5dc";
        public const string Fitness_center = "\ueb43";
        public const string Flag = "\ue153";
        public const string Flare = "\ue3e4";
        public const string Flash_auto = "\ue3e5";
        public const string Flash_off = "\ue3e6";
        public const string Flash_on = "\ue3e7";
        public const string Flight = "\ue539";
        public const string Flight_land = "\ue904";
        public const string Flight_takeoff = "\ue905";
        public const string Flip = "\ue3e8";
        public const string Flip_to_back = "\ue882";
        public const string Flip_to_front = "\ue883";
        public const string Folder = "\ue2c7";
        public const string Folder_open = "\ue2c8";
        public const string Folder_shared = "\ue2c9";
        public const string Folder_special = "\ue617";
        public const string Font_download = "\ue167";
        public const string Format_align_center = "\ue234";
        public const string Format_align_justify = "\ue235";
        public const string Format_align_left = "\ue236";
        public const string Format_align_right = "\ue237";
        public const string Format_bold = "\ue238";
        public const string Format_clear = "\ue239";
        public const string Format_color_fill = "\ue23a";
        public const string Format_color_reset = "\ue23b";
        public const string Format_color_text = "\ue23c";
        public const string Format_indent_decrease = "\ue23d";
        public const string Format_indent_increase = "\ue23e";
        public const string Format_italic = "\ue23f";
        public const string Format_line_spacing = "\ue240";
        public const string Format_list_bulleted = "\ue241";
        public const string Format_list_numbered = "\ue242";
        public const string Format_paint = "\ue243";
        public const string Format_quote = "\ue244";
        public const string Format_shapes = "\ue25e";
        public const string Format_size = "\ue245";
        public const string Format_strikethrough = "\ue246";
        public const string Format_textdirection_l_to_r = "\ue247";
        public const string Format_textdirection_r_to_l = "\ue248";
        public const string Format_underlined = "\ue249";
        public const string Forum = "\ue0bf";
        public const string Forward = "\ue154";
        public const string Forward_10 = "\ue056";
        public const string Forward_30 = "\ue057";
        public const string Forward_5 = "\ue058";
        public const string Free_breakfast = "\ueb44";
        public const string Fullscreen = "\ue5d0";
        public const string Fullscreen_exit = "\ue5d1";
        public const string Functions = "\ue24a";
        public const string G_translate = "\ue927";
        public const string Gamepad = "\ue30f";
        public const string Games = "\ue021";
        public const string Gavel = "\ue90e";
        public const string Gesture = "\ue155";
        public const string Get_app = "\ue884";
        public const string Gif = "\ue908";
        public const string Golf_course = "\ueb45";
        public const string Gps_fixed = "\ue1b3";
        public const string Gps_not_fixed = "\ue1b4";
        public const string Gps_off = "\ue1b5";
        public const string Grade = "\ue885";
        public const string Gradient = "\ue3e9";
        public const string Grain = "\ue3ea";
        public const string Graphic_eq = "\ue1b8";
        public const string Grid_off = "\ue3eb";
        public const string Grid_on = "\ue3ec";
        public const string Group = "\ue7ef";
        public const string Group_add = "\ue7f0";
        public const string Group_work = "\ue886";
        public const string Hd = "\ue052";
        public const string Hdr_off = "\ue3ed";
        public const string Hdr_on = "\ue3ee";
        public const string Hdr_strong = "\ue3f1";
        public const string Hdr_weak = "\ue3f2";
        public const string Headset = "\ue310";
        public const string Headset_mic = "\ue311";
        public const string Healing = "\ue3f3";
        public const string Hearing = "\ue023";
        public const string Help = "\ue887";
        public const string Help_outline = "\ue8fd";
        public const string High_quality = "\ue024";
        public const string Highlight = "\ue25f";
        public const string Highlight_off = "\ue888";
        public const string History = "\ue889";
        public const string Home = "\ue88a";
        public const string Hot_tub = "\ueb46";
        public const string Hotel = "\ue53a";
        public const string Hourglass_empty = "\ue88b";
        public const string Hourglass_full = "\ue88c";
        public const string Http = "\ue902";
        public const string Https = "\ue88d";
        public const string Image = "\ue3f4";
        public const string Image_aspect_ratio = "\ue3f5";
        public const string Import_contacts = "\ue0e0";
        public const string Import_export = "\ue0c3";
        public const string Important_devices = "\ue912";
        public const string Inbox = "\ue156";
        public const string Indeterminate_check_box = "\ue909";
        public const string Info = "\ue88e";
        public const string Info_outline = "\ue88f";
        public const string Input = "\ue890";
        public const string Insert_chart = "\ue24b";
        public const string Insert_comment = "\ue24c";
        public const string Insert_drive_file = "\ue24d";
        public const string Insert_emoticon = "\ue24e";
        public const string Insert_invitation = "\ue24f";
        public const string Insert_link = "\ue250";
        public const string Insert_photo = "\ue251";
        public const string Invert_colors = "\ue891";
        public const string Invert_colors_off = "\ue0c4";
        public const string Iso = "\ue3f6";
        public const string Keyboard = "\ue312";
        public const string Keyboard_arrow_down = "\ue313";
        public const string Keyboard_arrow_left = "\ue314";
        public const string Keyboard_arrow_right = "\ue315";
        public const string Keyboard_arrow_up = "\ue316";
        public const string Keyboard_backspace = "\ue317";
        public const string Keyboard_capslock = "\ue318";
        public const string Keyboard_hide = "\ue31a";
        public const string Keyboard_return = "\ue31b";
        public const string Keyboard_tab = "\ue31c";
        public const string Keyboard_voice = "\ue31d";
        public const string Kitchen = "\ueb47";
        public const string Label = "\ue892";
        public const string Label_outline = "\ue893";
        public const string Landscape = "\ue3f7";
        public const string Language = "\ue894";
        public const string Laptop = "\ue31e";
        public const string Laptop_chromebook = "\ue31f";
        public const string Laptop_mac = "\ue320";
        public const string Laptop_windows = "\ue321";
        public const string Last_page = "\ue5dd";
        public const string Launch = "\ue895";
        public const string Layers = "\ue53b";
        public const string Layers_clear = "\ue53c";
        public const string Leak_add = "\ue3f8";
        public const string Leak_remove = "\ue3f9";
        public const string Lens = "\ue3fa";
        public const string Library_add = "\ue02e";
        public const string Library_books = "\ue02f";
        public const string Library_music = "\ue030";
        public const string Lightbulb_outline = "\ue90f";
        public const string Line_style = "\ue919";
        public const string Line_weight = "\ue91a";
        public const string Linear_scale = "\ue260";
        public const string Link = "\ue157";
        public const string Linked_camera = "\ue438";
        public const string List = "\ue896";
        public const string Live_help = "\ue0c6";
        public const string Live_tv = "\ue639";
        public const string Local_activity = "\ue53f";
        public const string Local_airport = "\ue53d";
        public const string Local_atm = "\ue53e";
        public const string Local_bar = "\ue540";
        public const string Local_cafe = "\ue541";
        public const string Local_car_wash = "\ue542";
        public const string Local_convenience_store = "\ue543";
        public const string Local_dining = "\ue556";
        public const string Local_drink = "\ue544";
        public const string Local_florist = "\ue545";
        public const string Local_gas_station = "\ue546";
        public const string Local_grocery_store = "\ue547";
        public const string Local_hospital = "\ue548";
        public const string Local_hotel = "\ue549";
        public const string Local_laundry_service = "\ue54a";
        public const string Local_library = "\ue54b";
        public const string Local_mall = "\ue54c";
        public const string Local_movies = "\ue54d";
        public const string Local_offer = "\ue54e";
        public const string Local_parking = "\ue54f";
        public const string Local_pharmacy = "\ue550";
        public const string Local_phone = "\ue551";
        public const string Local_pizza = "\ue552";
        public const string Local_play = "\ue553";
        public const string Local_post_office = "\ue554";
        public const string Local_printshop = "\ue555";
        public const string Local_see = "\ue557";
        public const string Local_shipping = "\ue558";
        public const string Local_taxi = "\ue559";
        public const string Location_city = "\ue7f1";
        public const string Location_disabled = "\ue1b6";
        public const string Location_off = "\ue0c7";
        public const string Location_on = "\ue0c8";
        public const string Location_searching = "\ue1b7";
        public const string Lock = "\ue897";
        public const string Lock_open = "\ue898";
        public const string Lock_outline = "\ue899";
        public const string Looks = "\ue3fc";
        public const string Looks_3 = "\ue3fb";
        public const string Looks_4 = "\ue3fd";
        public const string Looks_5 = "\ue3fe";
        public const string Looks_6 = "\ue3ff";
        public const string Looks_one = "\ue400";
        public const string Looks_two = "\ue401";
        public const string Loop = "\ue028";
        public const string Loupe = "\ue402";
        public const string Low_priority = "\ue16d";
        public const string Loyalty = "\ue89a";
        public const string Mail = "\ue158";
        public const string Mail_outline = "\ue0e1";
        public const string Map = "\ue55b";
        public const string Markunread = "\ue159";
        public const string Markunread_mailbox = "\ue89b";
        public const string Memory = "\ue322";
        public const string Menu = "\ue5d2";
        public const string Merge_type = "\ue252";
        public const string Message = "\ue0c9";
        public const string Mic = "\ue029";
        public const string Mic_none = "\ue02a";
        public const string Mic_off = "\ue02b";
        public const string Mms = "\ue618";
        public const string Mode_comment = "\ue253";
        public const string Mode_edit = "\ue254";
        public const string Monetization_on = "\ue263";
        public const string Money_off = "\ue25c";
        public const string Monochrome_photos = "\ue403";
        public const string Mood = "\ue7f2";
        public const string Mood_bad = "\ue7f3";
        public const string More = "\ue619";
        public const string More_horiz = "\ue5d3";
        public const string More_vert = "\ue5d4";
        public const string Motorcycle = "\ue91b";
        public const string Mouse = "\ue323";
        public const string Move_to_inbox = "\ue168";
        public const string Movie = "\ue02c";
        public const string Movie_creation = "\ue404";
        public const string Movie_filter = "\ue43a";
        public const string Multiline_chart = "\ue6df";
        public const string Music_note = "\ue405";
        public const string Music_video = "\ue063";
        public const string My_location = "\ue55c";
        public const string Nature = "\ue406";
        public const string Nature_people = "\ue407";
        public const string Navigate_before = "\ue408";
        public const string Navigate_next = "\ue409";
        public const string Navigation = "\ue55d";
        public const string Near_me = "\ue569";
        public const string Network_cell = "\ue1b9";
        public const string Network_check = "\ue640";
        public const string Network_locked = "\ue61a";
        public const string Network_wifi = "\ue1ba";
        public const string New_releases = "\ue031";
        public const string Next_week = "\ue16a";
        public const string Nfc = "\ue1bb";
        public const string No_encryption = "\ue641";
        public const string No_sim = "\ue0cc";
        public const string Not_interested = "\ue033";
        public const string Note = "\ue06f";
        public const string Note_add = "\ue89c";
        public const string Notifications = "\ue7f4";
        public const string Notifications_active = "\ue7f7";
        public const string Notifications_none = "\ue7f5";
        public const string Notifications_off = "\ue7f6";
        public const string Notifications_paused = "\ue7f8";
        public const string Offline_pin = "\ue90a";
        public const string Ondemand_video = "\ue63a";
        public const string Opacity = "\ue91c";
        public const string Open_in_browser = "\ue89d";
        public const string Open_in_new = "\ue89e";
        public const string Open_with = "\ue89f";
        public const string Pages = "\ue7f9";
        public const string Pageview = "\ue8a0";
        public const string Palette = "\ue40a";
        public const string Pan_tool = "\ue925";
        public const string Panorama = "\ue40b";
        public const string Panorama_fish_eye = "\ue40c";
        public const string Panorama_horizontal = "\ue40d";
        public const string Panorama_vertical = "\ue40e";
        public const string Panorama_wide_angle = "\ue40f";
        public const string Party_mode = "\ue7fa";
        public const string Pause = "\ue034";
        public const string Pause_circle_filled = "\ue035";
        public const string Pause_circle_outline = "\ue036";
        public const string Payment = "\ue8a1";
        public const string People = "\ue7fb";
        public const string People_outline = "\ue7fc";
        public const string Perm_camera_mic = "\ue8a2";
        public const string Perm_contact_calendar = "\ue8a3";
        public const string Perm_data_setting = "\ue8a4";
        public const string Perm_device_information = "\ue8a5";
        public const string Perm_identity = "\ue8a6";
        public const string Perm_media = "\ue8a7";
        public const string Perm_phone_msg = "\ue8a8";
        public const string Perm_scan_wifi = "\ue8a9";
        public const string Person = "\ue7fd";
        public const string Person_add = "\ue7fe";
        public const string Person_outline = "\ue7ff";
        public const string Person_pin = "\ue55a";
        public const string Person_pin_circle = "\ue56a";
        public const string Personal_video = "\ue63b";
        public const string Pets = "\ue91d";
        public const string Phone = "\ue0cd";
        public const string Phone_android = "\ue324";
        public const string Phone_bluetooth_speaker = "\ue61b";
        public const string Phone_forwarded = "\ue61c";
        public const string Phone_in_talk = "\ue61d";
        public const string Phone_iphone = "\ue325";
        public const string Phone_locked = "\ue61e";
        public const string Phone_missed = "\ue61f";
        public const string Phone_paused = "\ue620";
        public const string Phonelink = "\ue326";
        public const string Phonelink_erase = "\ue0db";
        public const string Phonelink_lock = "\ue0dc";
        public const string Phonelink_off = "\ue327";
        public const string Phonelink_ring = "\ue0dd";
        public const string Phonelink_setup = "\ue0de";
        public const string Photo = "\ue410";
        public const string Photo_album = "\ue411";
        public const string Photo_camera = "\ue412";
        public const string Photo_filter = "\ue43b";
        public const string Photo_library = "\ue413";
        public const string Photo_size_select_actual = "\ue432";
        public const string Photo_size_select_large = "\ue433";
        public const string Photo_size_select_small = "\ue434";
        public const string Picture_as_pdf = "\ue415";
        public const string Picture_in_picture = "\ue8aa";
        public const string Picture_in_picture_alt = "\ue911";
        public const string Pie_chart = "\ue6c4";
        public const string Pie_chart_outlined = "\ue6c5";
        public const string Pin_drop = "\ue55e";
        public const string Place = "\ue55f";
        public const string Play_arrow = "\ue037";
        public const string Play_circle_filled = "\ue038";
        public const string Play_circle_outline = "\ue039";
        public const string Play_for_work = "\ue906";
        public const string Playlist_add = "\ue03b";
        public const string Playlist_add_check = "\ue065";
        public const string Playlist_play = "\ue05f";
        public const string Plus_one = "\ue800";
        public const string Poll = "\ue801";
        public const string Polymer = "\ue8ab";
        public const string Pool = "\ueb48";
        public const string Portable_wifi_off = "\ue0ce";
        public const string Portrait = "\ue416";
        public const string Power = "\ue63c";
        public const string Power_input = "\ue336";
        public const string Power_settings_new = "\ue8ac";
        public const string Pregnant_woman = "\ue91e";
        public const string Present_to_all = "\ue0df";
        public const string Print = "\ue8ad";
        public const string Priority_high = "\ue645";
        public const string Public = "\ue80b";
        public const string Publish = "\ue255";
        public const string Query_builder = "\ue8ae";
        public const string Question_answer = "\ue8af";
        public const string Queue = "\ue03c";
        public const string Queue_music = "\ue03d";
        public const string Queue_play_next = "\ue066";
        public const string Radio = "\ue03e";
        public const string Radio_button_checked = "\ue837";
        public const string Radio_button_unchecked = "\ue836";
        public const string Rate_review = "\ue560";
        public const string Receipt = "\ue8b0";
        public const string Recent_actors = "\ue03f";
        public const string Record_voice_over = "\ue91f";
        public const string Redeem = "\ue8b1";
        public const string Redo = "\ue15a";
        public const string Refresh = "\ue5d5";
        public const string Remove = "\ue15b";
        public const string Remove_circle = "\ue15c";
        public const string Remove_circle_outline = "\ue15d";
        public const string Remove_from_queue = "\ue067";
        public const string Remove_red_eye = "\ue417";
        public const string Remove_shopping_cart = "\ue928";
        public const string Reorder = "\ue8fe";
        public const string Repeat = "\ue040";
        public const string Repeat_one = "\ue041";
        public const string Replay = "\ue042";
        public const string Replay_10 = "\ue059";
        public const string Replay_30 = "\ue05a";
        public const string Replay_5 = "\ue05b";
        public const string Reply = "\ue15e";
        public const string Reply_all = "\ue15f";
        public const string Report = "\ue160";
        public const string Report_problem = "\ue8b2";
        public const string Restaurant = "\ue56c";
        public const string Restaurant_menu = "\ue561";
        public const string Restore = "\ue8b3";
        public const string Restore_page = "\ue929";
        public const string Ring_volume = "\ue0d1";
        public const string Room = "\ue8b4";
        public const string Room_service = "\ueb49";
        public const string Rotate_90_degrees_ccw = "\ue418";
        public const string Rotate_left = "\ue419";
        public const string Rotate_right = "\ue41a";
        public const string Rounded_corner = "\ue920";
        public const string Router = "\ue328";
        public const string Rowing = "\ue921";
        public const string Rss_feed = "\ue0e5";
        public const string Rv_hookup = "\ue642";
        public const string Satellite = "\ue562";
        public const string Save = "\ue161";
        public const string Scanner = "\ue329";
        public const string Schedule = "\ue8b5";
        public const string School = "\ue80c";
        public const string Screen_lock_landscape = "\ue1be";
        public const string Screen_lock_portrait = "\ue1bf";
        public const string Screen_lock_rotation = "\ue1c0";
        public const string Screen_rotation = "\ue1c1";
        public const string Screen_share = "\ue0e2";
        public const string Sd_card = "\ue623";
        public const string Sd_storage = "\ue1c2";
        public const string Search = "\ue8b6";
        public const string Security = "\ue32a";
        public const string Select_all = "\ue162";
        public const string Send = "\ue163";
        public const string Sentiment_dissatisfied = "\ue811";
        public const string Sentiment_neutral = "\ue812";
        public const string Sentiment_satisfied = "\ue813";
        public const string Sentiment_very_dissatisfied = "\ue814";
        public const string Sentiment_very_satisfied = "\ue815";
        public const string Settings = "\ue8b8";
        public const string Settings_applications = "\ue8b9";
        public const string Settings_backup_restore = "\ue8ba";
        public const string Settings_bluetooth = "\ue8bb";
        public const string Settings_brightness = "\ue8bd";
        public const string Settings_cell = "\ue8bc";
        public const string Settings_ethernet = "\ue8be";
        public const string Settings_input_antenna = "\ue8bf";
        public const string Settings_input_component = "\ue8c0";
        public const string Settings_input_composite = "\ue8c1";
        public const string Settings_input_hdmi = "\ue8c2";
        public const string Settings_input_svideo = "\ue8c3";
        public const string Settings_overscan = "\ue8c4";
        public const string Settings_phone = "\ue8c5";
        public const string Settings_power = "\ue8c6";
        public const string Settings_remote = "\ue8c7";
        public const string Settings_system_daydream = "\ue1c3";
        public const string Settings_voice = "\ue8c8";
        public const string Share = "\ue80d";
        public const string Shop = "\ue8c9";
        public const string Shop_two = "\ue8ca";
        public const string Shopping_basket = "\ue8cb";
        public const string Shopping_cart = "\ue8cc";
        public const string Short_text = "\ue261";
        public const string Show_chart = "\ue6e1";
        public const string Shuffle = "\ue043";
        public const string Signal_cellular_4_bar = "\ue1c8";
        public const string Signal_cellular_connected_no_internet_4_bar = "\ue1cd";
        public const string Signal_cellular_no_sim = "\ue1ce";
        public const string Signal_cellular_null = "\ue1cf";
        public const string Signal_cellular_off = "\ue1d0";
        public const string Signal_wifi_4_bar = "\ue1d8";
        public const string Signal_wifi_4_bar_lock = "\ue1d9";
        public const string Signal_wifi_off = "\ue1da";
        public const string Sim_card = "\ue32b";
        public const string Sim_card_alert = "\ue624";
        public const string Skip_next = "\ue044";
        public const string Skip_previous = "\ue045";
        public const string Slideshow = "\ue41b";
        public const string Slow_motion_video = "\ue068";
        public const string Smartphone = "\ue32c";
        public const string Smoke_free = "\ueb4a";
        public const string Smoking_rooms = "\ueb4b";
        public const string Sms = "\ue625";
        public const string Sms_failed = "\ue626";
        public const string Snooze = "\ue046";
        public const string Sort = "\ue164";
        public const string Sort_by_alpha = "\ue053";
        public const string Spa = "\ueb4c";
        public const string Space_bar = "\ue256";
        public const string Speaker = "\ue32d";
        public const string Speaker_group = "\ue32e";
        public const string Speaker_notes = "\ue8cd";
        public const string Speaker_notes_off = "\ue92a";
        public const string Speaker_phone = "\ue0d2";
        public const string Spellcheck = "\ue8ce";
        public const string Star = "\ue838";
        public const string Star_border = "\ue83a";
        public const string Star_half = "\ue839";
        public const string Stars = "\ue8d0";
        public const string Stay_current_landscape = "\ue0d3";
        public const string Stay_current_portrait = "\ue0d4";
        public const string Stay_primary_landscape = "\ue0d5";
        public const string Stay_primary_portrait = "\ue0d6";
        public const string Stop = "\ue047";
        public const string Stop_screen_share = "\ue0e3";
        public const string Storage = "\ue1db";
        public const string Store = "\ue8d1";
        public const string Store_mall_directory = "\ue563";
        public const string Straighten = "\ue41c";
        public const string Streetview = "\ue56e";
        public const string Strikethrough_s = "\ue257";
        public const string Style = "\ue41d";
        public const string Subdirectory_arrow_left = "\ue5d9";
        public const string Subdirectory_arrow_right = "\ue5da";
        public const string Subject = "\ue8d2";
        public const string Subscriptions = "\ue064";
        public const string Subtitles = "\ue048";
        public const string Subway = "\ue56f";
        public const string Supervisor_account = "\ue8d3";
        public const string Surround_sound = "\ue049";
        public const string Swap_calls = "\ue0d7";
        public const string Swap_horiz = "\ue8d4";
        public const string Swap_vert = "\ue8d5";
        public const string Swap_vertical_circle = "\ue8d6";
        public const string Switch_camera = "\ue41e";
        public const string Switch_video = "\ue41f";
        public const string Sync = "\ue627";
        public const string Sync_disabled = "\ue628";
        public const string Sync_problem = "\ue629";
        public const string System_update = "\ue62a";
        public const string System_update_alt = "\ue8d7";
        public const string Tab = "\ue8d8";
        public const string Tab_unselected = "\ue8d9";
        public const string Tablet = "\ue32f";
        public const string Tablet_android = "\ue330";
        public const string Tablet_mac = "\ue331";
        public const string Tag_faces = "\ue420";
        public const string Tap_and_play = "\ue62b";
        public const string Terrain = "\ue564";
        public const string Text_fields = "\ue262";
        public const string Text_format = "\ue165";
        public const string Textsms = "\ue0d8";
        public const string Texture = "\ue421";
        public const string Theaters = "\ue8da";
        public const string Thumb_down = "\ue8db";
        public const string Thumb_up = "\ue8dc";
        public const string Thumbs_up_down = "\ue8dd";
        public const string Time_to_leave = "\ue62c";
        public const string Timelapse = "\ue422";
        public const string Timeline = "\ue922";
        public const string Timer = "\ue425";
        public const string Timer_10 = "\ue423";
        public const string Timer_3 = "\ue424";
        public const string Timer_off = "\ue426";
        public const string Title = "\ue264";
        public const string Toc = "\ue8de";
        public const string Today = "\ue8df";
        public const string Toll = "\ue8e0";
        public const string Tonality = "\ue427";
        public const string Touch_app = "\ue913";
        public const string Toys = "\ue332";
        public const string Track_changes = "\ue8e1";
        public const string Traffic = "\ue565";
        public const string Train = "\ue570";
        public const string Tram = "\ue571";
        public const string Transfer_within_a_station = "\ue572";
        public const string Transform = "\ue428";
        public const string Translate = "\ue8e2";
        public const string Trending_down = "\ue8e3";
        public const string Trending_flat = "\ue8e4";
        public const string Trending_up = "\ue8e5";
        public const string Tune = "\ue429";
        public const string Turned_in = "\ue8e6";
        public const string Turned_in_not = "\ue8e7";
        public const string Tv = "\ue333";
        public const string Unarchive = "\ue169";
        public const string Undo = "\ue166";
        public const string Unfold_less = "\ue5d6";
        public const string Unfold_more = "\ue5d7";
        public const string Update = "\ue923";
        public const string Usb = "\ue1e0";
        public const string Verified_user = "\ue8e8";
        public const string Vertical_align_bottom = "\ue258";
        public const string Vertical_align_center = "\ue259";
        public const string Vertical_align_top = "\ue25a";
        public const string Vibration = "\ue62d";
        public const string Video_call = "\ue070";
        public const string Video_label = "\ue071";
        public const string Video_library = "\ue04a";
        public const string Videocam = "\ue04b";
        public const string Videocam_off = "\ue04c";
        public const string Videogame_asset = "\ue338";
        public const string View_agenda = "\ue8e9";
        public const string View_array = "\ue8ea";
        public const string View_carousel = "\ue8eb";
        public const string View_column = "\ue8ec";
        public const string View_comfy = "\ue42a";
        public const string View_compact = "\ue42b";
        public const string View_day = "\ue8ed";
        public const string View_headline = "\ue8ee";
        public const string View_list = "\ue8ef";
        public const string View_module = "\ue8f0";
        public const string View_quilt = "\ue8f1";
        public const string View_stream = "\ue8f2";
        public const string View_week = "\ue8f3";
        public const string Vignette = "\ue435";
        public const string Visibility = "\ue8f4";
        public const string Visibility_off = "\ue8f5";
        public const string Voice_chat = "\ue62e";
        public const string Voicemail = "\ue0d9";
        public const string Volume_down = "\ue04d";
        public const string Volume_mute = "\ue04e";
        public const string Volume_off = "\ue04f";
        public const string Volume_up = "\ue050";
        public const string Vpn_key = "\ue0da";
        public const string Vpn_lock = "\ue62f";
        public const string Wallpaper = "\ue1bc";
        public const string Warning = "\ue002";
        public const string Watch = "\ue334";
        public const string Watch_later = "\ue924";
        public const string Wb_auto = "\ue42c";
        public const string Wb_cloudy = "\ue42d";
        public const string Wb_incandescent = "\ue42e";
        public const string Wb_iridescent = "\ue436";
        public const string Wb_sunny = "\ue430";
        public const string Wc = "\ue63d";
        public const string Web = "\ue051";
        public const string Web_asset = "\ue069";
        public const string Weekend = "\ue16b";
        public const string Whatshot = "\ue80e";
        public const string Widgets = "\ue1bd";
        public const string Wifi = "\ue63e";
        public const string Wifi_lock = "\ue1e1";
        public const string Wifi_tethering = "\ue1e2";
        public const string Work = "\ue8f9";
        public const string Wrap_text = "\ue25b";
        public const string Youtube_searched_for = "\ue8fa";
        public const string Zoom_in = "\ue8ff";
        public const string Zoom_out = "\ue900";
        public const string Zoom_out_map = "\ue56b";
    }
}
