// not auto-generated. update every time you change icons.ttf!

#define ICON_MIN_FUR 0xe0f0
#define ICON_MAX_FUR 0xe164

// test
#define ICON_FUR_TEST0 u8"\ue0f0"
#define ICON_FUR_TEST1 u8"\ue0f1"
#define ICON_FUR_TEST2 u8"\ue0f2"
#define ICON_FUR_TEST3 u8"\ue0f3"
#define ICON_FUR_TEST4 u8"\ue0f4"

// instrument
#define ICON_FUR_INS_STD u8"\ue100"
#define ICON_FUR_INS_FM u8"\ue101"
#define ICON_FUR_INS_GB u8"\ue102"
#define ICON_FUR_INS_C64 u8"\ue103"
#define ICON_FUR_INS_AMIGA u8"\ue104"
#define ICON_FUR_INS_PCE u8"\ue105"
#define ICON_FUR_INS_AY u8"\ue106"
#define ICON_FUR_INS_AY8930 u8"\ue107"
#define ICON_FUR_INS_TIA u8"\ue108"
#define ICON_FUR_INS_SAA1099 u8"\ue109"
#define ICON_FUR_INS_VIC u8"\ue10a"
#define ICON_FUR_INS_PET u8"\ue10b"
#define ICON_FUR_INS_VRC6 u8"\ue10c"
#define ICON_FUR_INS_OPLL u8"\ue10d"
#define ICON_FUR_INS_OPL u8"\ue10e"
#define ICON_FUR_INS_FDS u8"\ue10f"
#define ICON_FUR_INS_VBOY u8"\ue110"
#define ICON_FUR_INS_N163 u8"\ue111"
#define ICON_FUR_INS_SCC u8"\ue112"
#define ICON_FUR_INS_OPZ u8"\ue113"
#define ICON_FUR_INS_POKEY u8"\ue114"
#define ICON_FUR_INS_BEEPER u8"\ue115"
#define ICON_FUR_INS_SWAN u8"\ue116"
#define ICON_FUR_INS_MIKEY u8"\ue117"
#define ICON_FUR_INS_VERA u8"\ue118"
#define ICON_FUR_INS_X1_010 u8"\ue119"
#define ICON_FUR_INS_VRC6_SAW u8"\ue11a"
#define ICON_FUR_INS_ES5506 u8"\ue11b"
#define ICON_FUR_INS_MULTIPCM u8"\ue11c"
#define ICON_FUR_INS_SNES u8"\ue11d"
#define ICON_FUR_INS_SU u8"\ue11e"
#define ICON_FUR_INS_NAMCO u8"\ue11f"
#define ICON_FUR_INS_OPL_DRUMS u8"\ue120"
#define ICON_FUR_INS_OPM u8"\ue121"
#define ICON_FUR_INS_NES u8"\ue122"
#define ICON_FUR_INS_MSM6258 u8"\ue123"
#define ICON_FUR_INS_MSM6295 u8"\ue124"
#define ICON_FUR_INS_ADPCMA u8"\ue125"
#define ICON_FUR_INS_ADPCMB u8"\ue126"
#define ICON_FUR_INS_SEGAPCM u8"\ue127"
#define ICON_FUR_INS_QSOUND u8"\ue128"
#define ICON_FUR_INS_YMZ280B u8"\ue129"
#define ICON_FUR_INS_RF5C68 u8"\ue12a"
#define ICON_FUR_INS_MSM5232 u8"\ue12b"
#define ICON_FUR_INS_T6W28 u8"\ue12c"
#define ICON_FUR_INS_K007232 u8"\ue12d"
#define ICON_FUR_INS_GA20 u8"\ue12e"
#define ICON_FUR_INS_POKEMINI u8"\ue12f"
#define ICON_FUR_INS_SM8521 u8"\ue130"
#define ICON_FUR_INS_PV1000 u8"\ue131"
#define ICON_FUR_INS_K053260 u8"\ue132"
#define ICON_FUR_INS_SCSP u8"\ue133"
#define ICON_FUR_INS_TED u8"\ue134"
#define ICON_FUR_INS_C140 u8"\ue135"
#define ICON_FUR_INS_C219 u8"\ue142"
#define ICON_FUR_INS_ESFM u8"\ue143"
#define ICON_FUR_INS_POWERNOISE u8"\ue15a"
#define ICON_FUR_INS_POWERNOISE_SAW u8"\ue15b"
#define ICON_FUR_INS_DAVE u8"\ue15c"
#define ICON_FUR_INS_NDS u8"\ue15d"
#define ICON_FUR_INS_GBA_DMA u8"\ue15e"
#define ICON_FUR_INS_GBA_MINMOD u8"\ue15f"
#define ICON_FUR_INS_BIFURCATOR u8"\ue160"  
#define ICON_FUR_INS_SID2 u8"\ue161"
#define ICON_FUR_INS_SUPERVISION u8"\ue162"
#define ICON_FUR_INS_UPD1771C u8"\ue163"
#define ICON_FUR_INS_SID3 u8"\ue164"

// sample editor
#define ICON_FUR_SAMPLE_APPLY_SILENCE u8"\ue136"
#define ICON_FUR_SAMPLE_FADEIN u8"\ue137"
#define ICON_FUR_SAMPLE_FADEOUT u8"\ue138"
#define ICON_FUR_SAMPLE_FILTER u8"\ue139"
#define ICON_FUR_SAMPLE_INSERT_SILENCE u8"\ue13a"
#define ICON_FUR_SAMPLE_INVERT u8"\ue13b"
#define ICON_FUR_SAMPLE_NORMALIZE u8"\ue13c"
#define ICON_FUR_SAMPLE_RESAMPLE u8"\ue13d"
#define ICON_FUR_SAMPLE_RESIZE u8"\ue13e"
#define ICON_FUR_SAMPLE_REVERSE u8"\ue13f"
#define ICON_FUR_SAMPLE_SIGN u8"\ue140"
#define ICON_FUR_SAMPLE_TRIM u8"\ue141"

// waveforms
#define ICON_FUR_NOISE u8"\ue14f"
#define ICON_FUR_PULSE1 u8"\ue150"
#define ICON_FUR_PULSE2 u8"\ue151"
#define ICON_FUR_PULSE3 u8"\ue152"
#define ICON_FUR_PULSE4 u8"\ue153"
#define ICON_FUR_PULSE5 u8"\ue154"
#define ICON_FUR_PULSE6 u8"\ue155"
#define ICON_FUR_PULSE7 u8"\ue156"
#define ICON_FUR_SAW u8"\ue157"
#define ICON_FUR_TRI u8"\ue158"

// misc
#define ICON_FUR_SINE u8"\ue144"
#define ICON_FUR_ADSR_A u8"\ue145"
#define ICON_FUR_ADSR_D u8"\ue146"
#define ICON_FUR_ADSR_S u8"\ue147"
#define ICON_FUR_ADSR_R u8"\ue148"
#define ICON_FUR_DEC_LINEAR u8"\ue149"
#define ICON_FUR_DEC_EXP u8"\ue14a"
#define ICON_FUR_INC_LINEAR u8"\ue14b"
#define ICON_FUR_INC_BENT u8"\ue14c"
#define ICON_FUR_VOL_DIRECT u8"\ue14d"
#define ICON_FUR_ADSR u8"\ue14e"
#define ICON_FUR_CROSSFADE u8"\ue159"
